import numpy as np
import scipy as sp
from typing import Tuple, Dict
from openfermion import QubitOperator
from qiskit.quantum_info import SparsePauliOp
from qiskit._accelerate.sparse_pauli_op import unordered_unique
import numba as nb

def matmul_GF2(A: np.array, B: np.array) -> np.array:
    """
    Matrix multiplication mod2, i.e. (A@B)%2. 
    
    Note this calls one of two numba functions. First one uses numpy dot functionality, the second
    does a manual multiplication over GF2 (that is safe for very large matrices)

    Args:
        A (np.array): boolean numpy array
        B (np.array): boolean numpy array
    Returns:
        matrix multiplication mod 2

    """
    if max(*A.shape, *B.shape)<2147483648: # 2**31
        return numba_dot_matmal_GF2(A,B)
    else:
        return numba_binary_matmal_GF2(A,B)

@nb.njit(fastmath=True, parallel=True, cache=True)
def numba_binary_matmal_GF2(A: np.array, B: np.array) -> np.array:
    """
    custom GF(2) multiplication using numba. i.e. C = (A@B) mod 2.
    
    Note function uses fact that multiplication over GF2 doesn't require sums for each matrix element in C
    instead it uses and AND operation (same as elementwise multiplicaiton of row,col pairs) 
    followed by XOR operation (same as taking sum of row,col multiplied pairs)
    
    Args:
        A (np.array): numpy boolean array
        B (np.array): numpy boolean array
    Returns:
        C (np.array): numpy boolean array of (A@B) mod 2
    """
    C = np.empty((A.shape[0], B.shape[1]), dtype=np.bool_)
    for i in nb.prange(C.shape[0]):
        for j in range(C.shape[1]):

            ## logical version 1 (slower) 
            # C[i, j] = bool(np.sum(np.logical_and(A[i, :], B[:, j]))%2)

            # # logical version 2 (slower) 
            # parity = False
            # for bit in np.logical_and(A[i, :], B[:, j]):
            #     parity^=bit
            # C[i, j] = parity

            ## faster loop
            acc = False
            for k in range(B.shape[0]):
                acc ^= A[i, k] & B[k, j]
            C[i, j] = acc
    return C

@nb.njit(fastmath=True, cache=True)
def numba_dot_matmal_GF2(A, B):
    """
    Matrix multiplication mod2, i.e. (A@B)%2. Note this function expects boolean input!

    Args:
        A (np.array): boolean numpy array
        B (np.array): boolean numpy array
    Returns:
        matrix multiplication mod 2

    """
    return np.asarray(np.dot(np.asarray(A, dtype = np.float64),
                             np.asarray(B, dtype = np.float64)
                             )%2, 
                      dtype = np.bool_)

def symplectic_to_string(symp_vec) -> str:
    """
    Returns string form of symplectic vector defined as (X | Z)

    Args:
        symp_vec (array): symplectic Pauliword array

    Returns:
        Pword_string (str): String version of symplectic array
    """
    n_qubits = len(symp_vec) // 2

    X_block = symp_vec[:n_qubits]
    Z_block = symp_vec[n_qubits:]

    Y_loc = np.logical_and(X_block, Z_block)
    X_loc = np.logical_xor(Y_loc, X_block)
    Z_loc = np.logical_xor(Y_loc, Z_block)

    char_aray = np.array(list('I' * n_qubits), dtype=str)

    char_aray[Y_loc] = 'Y'
    char_aray[X_loc] = 'X'
    char_aray[Z_loc] = 'Z'

    Pword_string = ''.join(char_aray)

    return Pword_string

def symplectic_to_openfermion(symp_vec, coeff) -> str:
    """Returns string form of symplectic vector defined as (X | Z).

    Args:
        symp_vec (array): symplectic Pauliword array

    Returns:
        Pword_string (str): String version of symplectic array
    """
    n_qubits = len(symp_vec) // 2

    X_block = symp_vec[:n_qubits]
    Z_block = symp_vec[n_qubits:]

    Y_loc = np.logical_and(X_block, Z_block)
    X_loc = np.logical_xor(Y_loc, X_block)
    Z_loc = np.logical_xor(Y_loc, Z_block)

    char_aray = np.array(list("I" * n_qubits), dtype=str)

    char_aray[Y_loc] = "Y"
    char_aray[X_loc] = "X"
    char_aray[Z_loc] = "Z"

    indices = np.array(range(n_qubits), dtype=str)
    char_aray = np.char.add(char_aray, indices)[np.where(char_aray != "I")[0]]

    Pword_string = " ".join(char_aray)

    return QubitOperator(Pword_string, coeff)

def string_to_symplectic(pauli_str, n_qubits):
    """
    Args:
        pauli_str (str): Pauli String to be converted to symplectic array
        n_qubits (int): Number of qubits

    Returns:
        symp_vec (array): symplectic Pauliword array
    """
    assert(len(pauli_str) == n_qubits), 'Number of qubits is incompatible with pauli string'
    assert (set(pauli_str).issubset({'I', 'X', 'Y', 'Z'})), 'pauliword must only contain X,Y,Z,I terms'

    char_aray = np.array(list(pauli_str), dtype=str)
    X_loc = (char_aray == 'X')
    Z_loc = (char_aray == 'Z')
    Y_loc = (char_aray == 'Y')

    symp_vec = np.zeros(2*n_qubits, dtype=int)
    symp_vec[:n_qubits] += X_loc
    symp_vec[n_qubits:] += Z_loc
    symp_vec[:n_qubits] += Y_loc
    symp_vec[n_qubits:] += Y_loc

    return symp_vec

def count1_in_int_bitstring(i):
    """
    Count number of "1" bits in integer i to be thought of in binary representation

    https://stackoverflow.com/questions/109023/how-to-count-the-number-of-set-bits-in-a-32-bit-integer#109025
    https://web.archive.org/web/20151229003112/http://blogs.msdn.com/b/jeuge/archive/2005/06/08/hakmem-bit-count.aspx

    Args:
        i (int):  Integer

    Returns:
        Number of "1" bits in the bitstring of integer 'i'.
    """
    i = i - ((i >> 1) & 0x55555555)  # add pairs of bits
    i = (i & 0x33333333) + ((i >> 2) & 0x33333333)  # quads
    return (((i + (i >> 4) & 0xF0F0F0F) * 0x1010101) & 0xffffffff) >> 24

def symplectic_to_sparse_matrix(symp_vec, coeff) -> sp.sparse.csr_matrix:
    """
    Returns (2**n x 2**n) matrix of paulioperator kronector product together
     defined from symplectic vector defined as (X | Z)

    This follows because tensor products of Pauli operators are one-sparse: they each have only
    one nonzero entry in each row and column

    Args:
        symp_vec (array): symplectic Pauliword array

    Returns:
        sparse_matrix (csr_matrix): sparse matrix of Pauliword
    """
    n_qubits = len(symp_vec) // 2

    X_block = symp_vec[:n_qubits]
    Z_block = symp_vec[n_qubits:]

    Y_number = sum(np.bitwise_and(X_block, Z_block).astype(int))
    global_phase = (-1j) ** Y_number

    # reverse order to match bitstring int valu of each bit in binary: [..., 32, 16, 8, 4, 2, 1]
    if n_qubits > 64:
        # numpy cannot handle ints over int64s (2**64) therefore use python objects
        binary_int_array = 1 << np.arange(n_qubits - 1, -1, -1).astype(object)
    else:
        binary_int_array = 1 << np.arange(n_qubits - 1, -1, -1)

    x_int = X_block @ binary_int_array
    z_int = Z_block @ binary_int_array

    dimension = 2**n_qubits

    row_ind = np.arange(dimension)
    col_ind = np.bitwise_xor(row_ind, x_int)

    row_inds_and_Zint = np.bitwise_and(row_ind, z_int)
    vals = global_phase * (-1) ** (count1_in_int_bitstring(row_inds_and_Zint)%2)

    sparse_matrix = sp.sparse.csr_matrix(
        (vals, (row_ind, col_ind)),
        shape=(dimension, dimension),
        dtype=complex
            )

    return coeff*sparse_matrix

def symplectic_cleanup(
        symp_matrix:    np.array, 
        coeff_vec:      np.array, 
        zero_threshold: float = None
    ) -> Tuple[np.array, np.array]:
    """ 
    Remove duplicated rows of symplectic matrix terms, whilst summing
    the corresponding coefficients of the deleted rows in coeff_vec

    Args:
        symp_matrix (np.array): Symplectic matrix.
        coeff_vec (np.array): Coefficient Vector.
        zero_threshold (float): Zero Threshold Value. By default it is set to 'None'.

    Returns: 
        Reduced symplectic matrix and reduced coefficient vector.
    """
    # # order lexicographically using a fast void view implementation...
    # # this scales to large numbers of qubits more favourably than np.lexsort
    # symp_matrix_view = np.ascontiguousarray(symp_matrix).view(
    #     np.dtype((np.void, symp_matrix.dtype.itemsize * symp_matrix.shape[1]))
    # )
    # re_order_indices = np.argsort(symp_matrix_view.ravel())
    # # sort the symplectic matrix and vector of coefficients accordingly
    # sorted_terms = symp_matrix[re_order_indices]
    # sorted_coeff = coeff_vec[re_order_indices]
    # # unique terms are those with non-zero entries in the adjacent row difference array
    # diff_adjacent = np.diff(sorted_terms, axis=0)
    # mask_unique_terms = np.append(True, np.any(diff_adjacent, axis=1))
    # reduced_symp_matrix = sorted_terms[mask_unique_terms]
    # # mask the term indices such that those which are skipped are summed under np.reduceat
    # summing_indices = np.arange(symp_matrix.shape[0])[mask_unique_terms]
    # reduced_coeff_vec = np.add.reduceat(sorted_coeff, summing_indices, axis=0)
    # # if a zero threshold is specified terms with sufficiently small coefficient will be dropped
    # if zero_threshold is not None:
    #     mask_nonzero = abs(reduced_coeff_vec)>zero_threshold
    #     reduced_symp_matrix = reduced_symp_matrix[mask_nonzero]
    #     reduced_coeff_vec = reduced_coeff_vec[mask_nonzero]

    # return reduced_symp_matrix, reduced_coeff_vec

    unique_locations, inverse_map = unordered_unique(symp_matrix.astype('uint16'))
    reduced_symp_matrix = symp_matrix[unique_locations]
    reduced_coeff_vec = np.zeros(unique_locations.shape[0], dtype=complex)
    np.add.at(reduced_coeff_vec, inverse_map, coeff_vec)
    if zero_threshold is not None:
        mask_nonzero = abs(reduced_coeff_vec)>zero_threshold
        reduced_symp_matrix = reduced_symp_matrix[mask_nonzero]
        reduced_coeff_vec = reduced_coeff_vec[mask_nonzero]
    return reduced_symp_matrix, reduced_coeff_vec

def random_symplectic_matrix(n_qubits,n_terms, diagonal=False, density=0.3):
    """ 
    Generates a random binary matrix of dimension (n_terms) x (2*n_qubits)
    Specifying diagonal=True will set the left hand side (X_block) to all zeros
    """
    if diagonal:
        Z_block = np.random.choice([True, False], size=[n_terms,n_qubits], p=[density/2, 1-density/2])
        return np.hstack([np.zeros_like(Z_block), Z_block])
    else:
        return np.random.choice([True, False], size=[n_terms,2*n_qubits], p=[density, 1-density])

def _rref_binary(matrix: np.array) -> np.array:
    """ 
    Row-reduced echelon form over the binary field (GF2) - rows are not reordered 
    here for efficiency (not required in some use cases, e.g. symmetry identification).

    Args:
        matrix(np.array): Matrix whose row-reduced echelon form over the binary field has to be found.

    Returns:
        Row-reduced echelon form over the binary field of input matrix.
    """
    rref_matrix = matrix.copy()
    # iterate over rows of array
    for i, row_i in enumerate(rref_matrix):
        # if not a row of zeros
        if np.any(row_i):
            # find the first non-zero entry of row i
            pivot = np.where(row_i)[0][0]
            # find the non-zero entries in column i
            update_set = np.setdiff1d(np.where(rref_matrix[:, pivot]), i)
            # XOR the rows containing non-zero entries in column i with row i
            rref_matrix[update_set] = np.bitwise_xor(rref_matrix[update_set], row_i)
            # the rows below i will now be zeroed out in the pivot column
    return rref_matrix

def rref_binary(matrix: np.array) -> np.array:
    """ 
    Full row-reduced echelon form with row reordering.

    Args:
        matrix(np.array): Matrix whose row-reduced echelon form with row reordering has to be found.

    Returns:
        Row-reduced echelon form of input matrix with row reordering of input matrix.
    """
    reduced = _rref_binary(matrix)
    row_order, col_order = zip(
        *sorted(
            [(i,np.where(row)[0][0]) for i,row in enumerate(reduced) if np.any(row)],
            key=lambda x:x[1]
        )
    )
    row_order = list(row_order) + list(set(range(reduced.shape[0])).difference(row_order))
    return reduced[row_order]

def _cref_binary(matrix: np.array) -> np.array:
    """ 
    Column-reduced echelon form with static columns (used in symmetry identification).
    
    Args:
        matrix(np.array): Matrix whose column-reduced echelon form with static columns has to be found.

    Returns:
        Column-reduced echelon form of input matrix with static columns.
    """
    return _rref_binary(matrix.T).T  

def cref_binary(matrix: np.array) -> np.array:
    """ 
    Column-reduced echelon form with ordered columns (used in basis reconstruction).
        
    Args:
        matrix(np.array): Matrix whose column-reduced echelon form with ordered columns has to be found.

    Returns:
        Column-reduced echelon form of input matrix with ordered columns.
    """
    return rref_binary(matrix.T).T        

def QubitOperator_to_dict(op: QubitOperator, num_qubits: int):
    """ 
    OpenFermion

    Args:
        op (QubitOperator): Qubit Operator
        num_qubits (int) : Number of qubiits.

    Returns:
        Dictionary format of Qubit Operator.
    """
    assert(type(op) == QubitOperator)
    op_dict = {}
    term_dict = op.terms
    terms = list(term_dict.keys())

    for t in terms:    
        letters = ['I' for i in range(num_qubits)]
        for i in t:
            letters[i[0]] = i[1]
        p_string = ''.join(letters)        
        op_dict[p_string] = term_dict[t]
         
    return op_dict

def SparsePauliOp_to_dict(op:SparsePauliOp) -> dict:
    """ 
    Qiskit
    
    Args:
        op (SparsePauliOp): Pauli Sum Operator

    Returns:
        Dictionary format of Pauli Sum Operator.
    """
    H_dict = {}
    for Pstr, coeff in op.to_list():
        H_dict[Pstr] = coeff
    return H_dict

def safe_PauliwordOp_to_dict(op) -> Dict[str, Tuple[float, float]]:
    """ 
    Stores the real and imaginary parts of the coefficient separately in a tuple.

    Args:
        op (PauliwordOp): Weighted linear combination of N-fold Pauli operators.
    Returns:
        dict_out (dict): Dictionary of the form {pstring:(real, imag)}.
    """
    terms, coeffs = zip(*op.to_dictionary.items())
    coeffs = [(n.real, n.imag) for n in coeffs]
    dict_out = dict(zip(terms, coeffs))
    return dict_out

def safe_QuantumState_to_dict(psi) -> Dict[str, Tuple[float, float]]:
    """ 
    Stores the real and imaginary parts of the coefficient separately in a tuple.

    Args:
        op (QuantumState): Weighted linear combination of N-fold Pauli operators.
    Returns:
        dict_out (dict): Dictionary of the form {pstring:(real, imag)}.
    """
    terms, coeffs = zip(*psi.to_dictionary.items())
    coeffs = [(n.real, n.imag) for n in coeffs]
    dict_out = dict(zip(terms, coeffs))
    return dict_out

def mul_symplectic(
        symp_vec1: np.array,
        coeff1: complex,
        symp_vec2: np.array,
        coeff2: complex) -> Tuple[np.array, complex, int]:
    """ 
    Performs Pauli multiplication with phases at the level of the symplectic
    vector (1D here!). The phase compensation is implemented as per https://doi.org/10.1103/PhysRevA.68.042318.

    P1 * P2 is performed

    Args:
        symp_vec1 (np.array) : 1D vector of left Pauli operator
        coeff1 (float): coefficient of left  Pauli operator

        symp_vec2 (np.array) : 1D vector of right Pauli operator
        coeff2 (float): coefficient of right  Pauli operator

    Returns:
        output_symplectic_vec (np.array): binary symplectic output (Pauli opertor out)
        coeff_vec (complex): complex coeff with correct phase
    """
    X_block1, Z_block1 = np.split(symp_vec1, 2)
    X_block2, Z_block2 = np.split(symp_vec2, 2)

    # number of Y terms in left Pauli operator
    Y_count1 = np.sum(np.bitwise_and(X_block1, Z_block1), axis=0)
    # number of Y terms in right Pauli operator
    Y_count2 = np.sum(np.bitwise_and(X_block2, Z_block2), axis=0)

    # phaseless multiplication is binary addition in symplectic representation
    output_symplectic_vec = np.bitwise_xor(symp_vec1, symp_vec2)
    # phase is determined by Y counts plus additional sign flip
    Y_count_out = np.sum(np.bitwise_and(*np.split(output_symplectic_vec, 2)), axis=0) # number of Y terms in output
    # X_block of first op and Z_block of second op
    sign_change = (-1) ** (
            np.sum(np.bitwise_and(X_block1, Z_block2), axis=0) % 2
    )  # mod 2 as only care about parity
    # final phase modification
    phase_mod = sign_change * (1j) ** ((3 * (Y_count1 + Y_count2) + Y_count_out) % 4)  # mod 4 as roots of unity
    coeff_vec = phase_mod * coeff1 * coeff2
    return output_symplectic_vec, coeff_vec #, Y_count_out

def unit_n_sphere_cartesian_coords(angles: np.array) -> np.array:
    """ 
    Input an array of angles of length n, returns the n+1 cartesian coordinates 
    of the corresponding unit n-sphere in (n+1)-dimensional Euclidean space.

    Args:
        angles (np.array): Array of angles of length n of the corresponding unit n-sphere in (n+1)-dimensional Euclidean space.

    Returns:
        Numpy Array of n+1 cartesian coordinates.
    """
    cartesians = [np.prod(np.sin(angles[:i]))*np.cos(angles[i]) for i in range(len(angles))]
    cartesians.append(np.prod(np.sin(angles)))
    return np.array(cartesians)

def binomial_coefficient(n,k):
    """ 
    Calculate the binomial coefficient "n choose k" or denoted as "C(n, k)," represents the number of ways to choose k objects from a set of n objects without considering their order.
    Differs from np.math.comb as this allows non-integer n.

    Args: 
        n: Total number of objects.
        k: Number of object to choose from a set of n objects.

    Returns:
        Binomial coefficient "n choose k".
    """
    prod = 1
    for r in range(k):
        prod *= (n-r)/(k-r) 
    return prod

def check_independent(operators):
    """
    Check if the input PauliwordOp contains algebraically dependent terms.
    
    Args:
        operators (PauliwordOp): Operators.

    Returns:
        Returns True, if input operators contains algebraically dependent terms.
    """
    if operators.n_terms>2*operators.n_qubits:
        # cannot have indp. set of Pauli operators of more than 2N terms
        return False
    else:
        ind_check = _rref_binary(operators.symp_matrix)
        return ~np.any(np.all(~ind_check, axis=1))

def check_jordan_independent(operators):
    """ 
    Check if the input PauliwordOp contains algebraically dependent terms under jordan product
    (note input can be noncontextual, but contain dependent terms!)

    Args:
        operators (PauliwordOp): Operators.
    
    Returns:
        Returns True, if input operators contains algebraically dependent terms.

    
    test with : H ={
                     'IIIZ': (1+0j),
                     'IIZI': (1+0j),
                     'ZIII': (1+0j),
                     'IXII': (1+0j),
                     'XIIX': (1+0j)
                     }
    # clique =  [IIIZ, XIIX]
    # Z2     =  [IIZI, ZIIZ, IXII]

    """
    if operators.n_terms> 3*operators.n_qubits:
        # cannot have a jordan indp. set larger than 3N
        ## (e.g. only SINGLE X,Y,Z on every qubit position)
        ## {XI, ZI, YI, IX, IY, IZ}
        return False

    # check if globally commuting subset of operators is independent
    comm_mask = np.sum(operators.commutes_termwise(operators), axis=1) == operators.n_terms
    comm_part = operators[comm_mask]
    if not check_independent(comm_part):
        # commuting subset is not independent
        ## {YY, XX, ZZ} is example where this would fail! and is why we check here
        ## as would pass test below
        return False
    
    ## Final check over WHOLE operator to see if jordan indp. (not need to include commuting part! See unit test)
    # do row reduction where Y terms treated seperately
    Y_block = np.logical_and(operators.Z_block, operators.X_block)
    XZY_block = np.hstack((np.logical_xor(operators.X_block, Y_block), np.hstack((np.logical_xor(operators.Z_block, Y_block), Y_block))))
    ind_check = _rref_binary(XZY_block)
    return ~np.any(np.all(~ind_check, axis=1))


def check_adjmat_noncontextual(adjmat) -> bool:
    """
    Check whether the input boolean square matrix has a noncontextual structure...
    ... see https://doi.org/10.1103/PhysRevLett.123.200501 for details.

    Args:
        adjmat: Boolean square matrix.

    Returns:
        Returns True, if input matrix has a noncontextual structure.
    """
    # mask the terms that do not commute universally amongst the operator
    mask_non_universal = np.where(~np.all(adjmat, axis=1))[0]
    # look only at the unique rows in the masked adjacency matrix -
    # identical rows correspond with operators of the same clique
    unique_commutation_character = np.unique(
        adjmat[mask_non_universal,:][:,mask_non_universal],
        axis=0
    )
    # if the unique commutation characteristics are disjoint, i.e. no overlapping ones 
    # between rows, the operator is noncontextual - hence we sum over rows and check
    # the resulting vector consists of all ones.
    return np.all(np.count_nonzero(unique_commutation_character, axis=0)==1)


def perform_noncontextual_sweep(operator) -> "PauliwordOp":
    """
    Given an ordered operator, sweep over its terms once in order keeping terms that are noncontextual

    Args:
        operator (PauliwordOp): Ordered operator.

    Returns:
        List of terms maintaining noncontextuality.
    """

    # initialize noncontextual operator with first element of input operator
    noncon_indices = np.array([0])
    adjmat = np.array([[True]], dtype=bool)
    for index, term in enumerate(operator[1:]):
        # pad the adjacency matrix term-by-term - avoids full construction each time
        adjmat_vector = np.append(term.commutes_termwise(operator[noncon_indices]), True)
        adjmat_padded = np.pad(adjmat, pad_width=((0, 1), (0, 1)), mode='constant')
        adjmat_padded[-1,:] = adjmat_vector; adjmat_padded[:,-1] = adjmat_vector
        # check whether the adjacency matrix has a noncontextual structure
        if check_adjmat_noncontextual(adjmat_padded):
            noncon_indices = np.append(noncon_indices, index+1)
            adjmat = adjmat_padded

    return operator[noncon_indices] 

def binary_array_to_int(bin_arr):
    """
    Function to convert an array composed of rows of binary into integers.

    Args:
        bin_arr(np.array): 2D numpy array of binary.
    Returns:
        int_arr (np.array): 1D numpy array of ints.
    """

    if bin_arr.shape[1] < 64:
        b2i = 2 ** np.arange(bin_arr.shape[1] - 1, -1, -1)
    else:
        b2i = 2 ** np.arange(bin_arr.shape[1] - 1, -1, -1, dtype=float)
        # b2i = 2 ** np.arange(bin_arr.shape[1] - 1, -1, -1, dtype=object)

    int_arr = np.einsum('j, ij->i', b2i, bin_arr)
    # int_arr = (b2i*bin_arr).sum(axis=1)

    ## slower as does matrix product rather than multiplication along rows followed by a sum!
    # int_arr = bin_arr @ b2i

    return int_arr


import numpy as np
from copy import deepcopy
from scipy.optimize import differential_evolution
from symmer.operators import PauliwordOp, IndependentOp
from typing import Union, Optional
from symmer.utils import random_anitcomm_2n_1_PauliwordOp, product_list

def norm(vector: np.array) -> float:
    """
    Args:
        vector (np.array): Vector whose 12-norm has to be found.

    Returns:
        l2-norm of input vector
    """
    return np.sqrt(np.dot(vector, vector.conjugate()))

def lp_norm(vector: np.array, p:int=2) -> float:
    """
    Args:
        vector (np.array): Vector whose lp-norm has to be found.
        p (int): Power.  It is the power to which the absolute value of elements are raised before summation. is a positive real number.

    Returns:
        lp-norm of vector
    """
    return np.power(np.sum(np.power(np.abs(vector), p)), 1/p)

def one_qubit_noncontextual_gs(op: PauliwordOp):
    assert op.n_qubits == 1, 'Operator consists of more than one qubit'
    op.to

def basis_score(
        weighting_operator: PauliwordOp,
        basis: IndependentOp,
        p:int=1
    ) -> float:
    """ 
    Evaluate the score of an input basis according 
    to the basis weighting operator, for example:
        - set Hamiltonian cofficients to 1 for unweighted number of commuting terms
        - specify as the SOR Hamiltonian to weight according to second-order response
        - input UCC operator to weight according to coupled-cluster theory <- best performance
        - if None given then weights by Hamiltonian coefficient magnitude
    
    p determines which norm is used, i.e. lp --> (\sum_{t} |t|^p)^(1/p)

    Args:
        weighting_operator (PauliwordOp): Basis weighting operator.
        basis (IndependentOp): Basis whose basis score has to be determined.
        p (int): Power which determines which norm is used. Its default value is 1.

    Returns:
        Basis score (float) of the input basis.
    """
    # mask terms of the weighting operator that are preserved under projection over the basis
    mask_preserved = np.where(np.all(weighting_operator.commutes_termwise(basis),axis=1))[0]
    return (
        lp_norm(weighting_operator.coeff_vec[mask_preserved], p=p) /
        lp_norm(weighting_operator.coeff_vec, p=p)
    )

def update_eigenvalues(
        generators: IndependentOp, 
        stabilizers: IndependentOp
    ) -> None:
    """ 
    Update the +/-1 eigenvalue assigned to the input stabilizer according to the noncontextual ground state configuration.
    
    Args:
        generators (IndependentOp): Generator
        stabilizers (IndependentOp): Stabilizer 
    """
    reconstruction, successfully_reconstructed = stabilizers.generator_reconstruction(generators)
    if ~np.all(successfully_reconstructed):
        raise ValueError('Generators not sufficient to reconstruct symmetry operators')
    stabilizers.coeff_vec = (-1) ** np.count_nonzero(
        np.bitwise_and(
            reconstruction, 
            np.asarray(generators.coeff_vec)==-1
        ),
        axis=1
    )

class StabilizerIdentification:
    def __init__(self,
        weighting_operator: PauliwordOp,
        use_X_only = False
        ) -> None:
        """
        Args:
            weighting_operator (PauliwordOp): Basis weighting operator. By default, it is set to None.
            use_X_only (bool): Default value is 'False'.
        """
        self.weighting_operator = weighting_operator
        self.use_X_only = use_X_only
        self.build_basis_weighting_operator()

    def build_basis_weighting_operator(self):
        if self.use_X_only:
            X_block = self.weighting_operator.X_block
            self.weighting_operator = PauliwordOp(
                np.hstack([X_block, np.zeros_like(X_block)]), 
                np.abs(self.weighting_operator.coeff_vec)
            ).cleanup()
        self.basis_weighting = self.weighting_operator.sort(by='magnitude')
        self.qubit_positions = np.arange(self.weighting_operator.n_qubits)
        self.term_region = [0,self.basis_weighting.n_terms]
        
    def symmetry_generators_by_term_significance(self, n_preserved):
        """ 
        Set the number of terms to be preserved in order of coefficient magnitude,
        Then generate the largest symmetry basis that preserves them.

        Args: 
            n_preserved (int): Number of terms to be preserved in order of coefficient magnitude.
        
        Returns:
            The largest symmetry basis that preserves order of coefficient magnitude.
        """
        preserve = self.basis_weighting[:n_preserved]
        stabilizers = IndependentOp.symmetry_generators(preserve, commuting_override=True)
        mask_diag = np.where(~np.any(stabilizers.X_block, axis=1))[0]
        return IndependentOp(stabilizers.symp_matrix[mask_diag], stabilizers.coeff_vec[mask_diag])

    def symmetry_generators_by_subspace_dimension(self, n_sim_qubits, region=None):
        """
        Args:
            n_sim_qubits (int): Number of qubits to simulate.
            region (list[int]): Region

        Returns:
            Symetry generators by subspace dimension.
        """
        if region is None:
            region = deepcopy(self.term_region)
        assert(n_sim_qubits < self.basis_weighting.n_qubits), 'Number of qubits to simulate exceeds those in the operator'
        assert(region[1]-region[0]>1), 'Search region collapsed without identifying any stabilizers'

        n_terms = sum(region)//2
        stabilizers = self.symmetry_generators_by_term_significance(n_terms)
        current_n_qubits = self.basis_weighting.n_qubits - stabilizers.n_terms
        sign = np.sign(current_n_qubits - n_sim_qubits)

        if sign==0:
            # i.e. n_sim_qubits == current_n_qubits
            return stabilizers
        elif sign==+1:
            # i.e. n_sim_qubits < current_n_qubits
            region[1] = n_terms
        else:
            region[0] = n_terms
            
        return self.symmetry_generators_by_subspace_dimension(n_sim_qubits, region=region)

class ObservableBiasing:
    """ 
    Class for re-weighting Hamiltonian terms based on some criteria, such as HOMO-LUMO bias.
    
    Attributes:
        HOMO_bias (float): HUMO Bias. Its value is in between 0 and 1. By default it's value is set to be 0.2
        LUMO_bias (float): LUMO Bias. Its value is in between 0 and 1. By default it's value is set to be 0.2
        seperation (int):  Separation between the two distributions. By default it is set to 1. A value of 1 means each distribution is peaked either side of the HOMO-LUMO gap.
    """
    # HOMO/LUMO bias is a value between 0 and 1 representing how sharply 
    # peaked the Gaussian distributions centred at each point should be
    HOMO_bias = 0.2
    LUMO_bias = 0.2
    # Can also specify the separation between the two distributions... 
    # a value of 1 means each is peaked either side of the HOMO-LUMO gap
    separation = 1
    
    def __init__(self, base_operator: PauliwordOp, HOMO_LUMO_gap) -> None:
        """
        Args:
            base_operator (PauliwordOp): Base Operator.
            HOMO_LUMO_gap: HOMO-LUMO gap. It should be specified as the mid-point between the HOMO and LUMO indices.
        """
        self.base_operator = base_operator
        assert(
            HOMO_LUMO_gap - int(HOMO_LUMO_gap) == 0.5
        ), 'HOMO_LUMO_gap should be specified as the mid-point between the HOMO and LUMO indices'
        self.HOMO_LUMO_gap = HOMO_LUMO_gap
        # shift qubit positions such that HOMO-LUMO gap is centred at zero
        self.shifted_q_pos = np.arange(base_operator.n_qubits) - self.HOMO_LUMO_gap
        
    def HOMO_LUMO_bias_curve(self) -> np.array:
        """ 
        Curve constructed from two gaussians centred either side of the HOMO-LUMO gap.
        The standard deviation for each distribution can be tuned independently via
        the parameters HOMO_sig (lower population), LUMO_sig (upper population) in [0, pi/2].

        Returns:
            HOMO LUMO bias curve (np.array).
        """
        shift = self.separation - 1/2
        # standard deviation about the HOMO/LUMO-centred Gaussian distributions:
        HOMO_sigma = np.tan((1-self.HOMO_bias)*np.pi/2) 
        LUMO_sigma = np.tan((1-self.LUMO_bias)*np.pi/2)
        # lower population (centred at HOMO)
        if HOMO_sigma!=0:
            L = np.exp(-np.square((self.shifted_q_pos+shift)/HOMO_sigma)/2)
        else:
            non_zero_index = int(self.HOMO_LUMO_gap-shift)
            L = np.eye(1,self.base_operator.n_qubits,non_zero_index).reshape(self.base_operator.n_qubits)
        # upper population (centred at LUMO)
        if LUMO_sigma!=0:
            U = np.exp(-np.square((self.shifted_q_pos-shift)/LUMO_sigma)/2)
        else:
            non_zero_index = int(self.HOMO_LUMO_gap+shift)
            U = np.eye(1,self.base_operator.n_qubits,non_zero_index).reshape(self.base_operator.n_qubits)
        return (L + U)/2
    
    def HOMO_LUMO_biased_operator(self) -> np.array:
        """ 
        - First converts the base operator to a PauliwordOp consisting of Pauli I, X 
            (since only interested in where the terms can affect orbital occupation)
        - Second, assigns a weight to each nontrivial qubit position according to the bias curve
            and sums the total HOM-LUMO-biased contribution. This is multiplied by the coefficient
            vector to re-weight according to how close to the gap each term acts.

        Returns:
            reweighted_operator (PauliwordOp): Reweighted Operator.
        """
        reweighted_operator = self.base_operator.copy()
        reweighted_operator.coeff_vec = np.sum(
            reweighted_operator.X_block*self.HOMO_LUMO_bias_curve(), 
            axis=1
        )*reweighted_operator.coeff_vec
        return reweighted_operator

def stabilizer_walk(
        n_sim_qubits,
        biasing_operator: ObservableBiasing,
        weighting_operator: PauliwordOp = None,
        print_info: bool = False,
        use_X_only: bool = False
    ) -> IndependentOp:
    """
    Args:
        n_sim_qubits (int): Number of qubits to simulate.
        biasing_operator (ObservableBiasing): Biasing Operator.
        weighting_operator (PauliwordOp): Basis weighting operator. By default, it is set to None.
        print_info (bool): If True, Info about optimal score for HUMO/LUMO bias is printed. By default, it is set to 'False'.
        use_X_only:  Default value is 'False'.

    Returns:
        S IndependentOp: Stablizers
    """
    if weighting_operator is None:
        weighting_operator = biasing_operator.base_operator
        
    def get_stabilizers(x):
        biasing_operator.HOMO_bias,biasing_operator.LUMO_bias = x
        biased_op = biasing_operator.HOMO_LUMO_biased_operator()
        stabilizers = StabilizerIdentification(biased_op, use_X_only=use_X_only)
        S = stabilizers.symmetry_generators_by_subspace_dimension(n_sim_qubits)
        return(S)
    
    def objective(x):
        S = get_stabilizers(x)
        stab_score = basis_score(weighting_operator, S)
        return -stab_score
    
    opt_out = differential_evolution(objective, bounds=[(0,1),(0,1)])
    stab_score =-opt_out['fun']
    bias_param =opt_out['x']
    S = get_stabilizers(bias_param)
    
    if print_info:
        print(f'Optimal score w(S)={stab_score} for HOMO/LUMO bias {bias_param}')
    
    return S

def get_noncon_generators_from_commuting_stabilizers(stabilizers: Union[PauliwordOp, IndependentOp],
                                                     weighting_operator: PauliwordOp,
                                                     return_clique_only: Optional[bool] = False) -> IndependentOp:
    """
    Given a set of commuting stabilizers and weighting operator find best noncontextual generating set
    (ie works out best anticommuting addition to generators that reconstructs most of the weighting_operator)

    Args:
        stabilizers (PauliwordOp): operator containing commuting symmetries
        weighting_operator (PauliwordOp): operator to inform ac choice

    Returns:
        new_stabilizers (IndependentOp): noncontextual generators that contain ac component (generates noncon op under Jordan product)
    """
    if not np.all(stabilizers.commutes_termwise(stabilizers)):
        # stabilizers already contain ac component
        return stabilizers #, PauliwordOp.empty(stabilizers.n_qubits).cleanup()
    else:
        # below generates the generators of inout stabilizers
        generators = stabilizers.generators

    best_l1_norm = -1
    # find qubits uniquely defined by generators
    unique_q_inds = ~(np.sum(np.logical_xor(generators.Z_block, generators.X_block), axis=0)-1).astype(bool)
    for stab in generators:
        # find unique non identity positions
        act_positions = np.logical_and(np.logical_xor(stab.Z_block, stab.X_block)[0], unique_q_inds)

        # work out number of qubits on these positions
        n_act_qubits = np.sum(act_positions)

        # find AC clique of size 2n containing given stabilizer
        ac_basis = random_anitcomm_2n_1_PauliwordOp(n_act_qubits, apply_clifford=False)[1:]
        new_basis = PauliwordOp(np.zeros((n_act_qubits * 2, stab.n_qubits * 2), dtype=bool),
                                np.ones(n_act_qubits * 2))

        new_basis.symp_matrix[:, [*act_positions, *act_positions]] = ac_basis.symp_matrix

        # ensure stab is in new_basis
        gen, mask = stab.generator_reconstruction(new_basis)
        required_products = gen[0].nonzero()[0][1:]
        if len(required_products) > 0:
            prod = product_list(new_basis[required_products])
            new_basis = (new_basis * prod).cleanup()
        new_basis.coeff_vec = np.ones_like(new_basis.coeff_vec)

        # find best reconstruction
        _, mask = weighting_operator.generator_reconstruction(new_basis)
        success = weighting_operator[mask]
        l1_norm = np.linalg.norm(success.coeff_vec, ord=1)

        if l1_norm > best_l1_norm:
            new_stabilizers = generators - stab + new_basis
            best_l1_norm = l1_norm
            stab_used = stab.copy()

    assert new_stabilizers.is_noncontextual, 'new stabilizers are not noncontextual'

    # commuting_stabs = IndependentOp.from_PauliwordOp(stabilizers)
    # anticommuting_stabs = IndependentOp.from_PauliwordOp(new_stabilizers) - commuting_stabs
    # return commuting_stabs, anticommuting_stabs
    if return_clique_only:
        return IndependentOp.from_PauliwordOp(new_stabilizers) - generators, stab_used
    else:
        return IndependentOp.from_PauliwordOp(new_stabilizers)


"""init for evolution."""
from .exponentiation import trotter
from .gate_library import *
from .utils import get_CNOT_connectivity_graph, topology_match_score
from .decomposition import qasm_to_PauliwordOp, PauliwordOp_to_QuantumCircuit
from .circuit_symmerlator import CircuitSymmerlator
from .variational_optimization import VQE_Driver, ADAPT_VQE


from symmer.operators import PauliwordOp
import numpy as np
from typing import Dict, List, Optional, Tuple, Union
import warnings

warnings.simplefilter('always', UserWarning)
class AntiCommutingOp(PauliwordOp):

    def __init__(self,
                 AC_op_symp_matrix: np.array,
                 coeff_list: np.array):
        """
        Args:
            AC_op_symp_matrix (np.array): The symmetric matrix representation of the anti-commuting operator.
            coeff_list (np.array): The coefficient list associated with the anti-commuting operator.

        Raises:
            AssertionError: If the operators do not anticommute.
        """
        super().__init__(AC_op_symp_matrix, coeff_list)

        # check all operators anticommute
        adj_mat = self.adjacency_matrix
        adj_mat[np.diag_indices_from(adj_mat)] = False
        assert ~np.any(adj_mat), 'operator needs to be made of anti-commuting Pauli operators'

        self.X_sk_rotations = []
        self.R_LCU = None

    @classmethod
    def from_list(cls,
            pauli_terms :List[str],
            coeff_vec:   List[complex] = None
        ) -> "AntiCommutingOp":
        """
        Args:
            pauli_terms (List[str]): A list of Pauli terms represented as strings.
            coeff_vec (List[complex], optional): A list of complex coefficients associated with the Pauli terms.
                If not provided, the default coefficients are assumed to be 1.0 for each term.

        Returns:
            AntiCommutingOp: An AntiCommutingOp instance initialized from the list of Pauli terms.
        """
        PwordOp = super().from_list(pauli_terms, coeff_vec)
        return cls.from_PauliwordOp(PwordOp)

    @classmethod
    def from_dictionary(cls,
            operator_dict: Dict[str, complex]
        ) -> "AntiCommutingOp":
        """ 
        Initialize a PauliwordOp from its dictionary representation {pauli:coeff, ...}

        Args:
            operator_dict (Dict[str, complex]): A dictionary representing the operator, where the keys are Pauli strings
                and the values are complex coefficients.

        Returns:
            AntiCommutingOp: An AntiCommutingOp instance initialized from the dictionary representation.
        """
        PwordOp = super().from_dictionary(operator_dict)
        return cls.from_PauliwordOp(PwordOp)

    @classmethod
    def from_PauliwordOp(cls,
            PwordOp: PauliwordOp
        ) -> 'AntiCommutingOp':
        """
        Args:
            PwordOp (PauliwordOp): The PauliwordOp instance to initialize the AntiCommutingOp from.

        Returns:
            AntiCommutingOp: An AntiCommutingOp instance initialized from the given PauliwordOp.
        """
        return cls(PwordOp.symp_matrix, PwordOp.coeff_vec)


    def get_least_dense_term_index(self):
        """
        Takes the current symp_matrix of object and finds the index of the least dense Pauli
        operator (aka least Pauli matrices). This can be used to define the term to reduce too

        Note one needs to re-run this function if ordering changed (e.g. if lexicographical_sort is run)

        Returns:
            s_index (int): index of least dense term in objects symp_matrix and coeff_vec.
        """
        ### np.logical_or(X_block, Z_block)
        # pos_terms_occur = np.logical_or(self.symp_matrix[:, :self.n_qubits], self.symp_matrix[:, self.n_qubits:])
        # int_list = pos_terms_occur @ (1 << np.arange(pos_terms_occur.shape[1], dtype=object)[::-1])
        # s_index = np.argmin(int_list)

        pos_terms_occur = np.logical_or(self.symp_matrix[:, :self.n_qubits], self.symp_matrix[:, self.n_qubits:])
        symp_matrix_view = np.ascontiguousarray(pos_terms_occur).view(
            np.dtype((np.void, pos_terms_occur.dtype.itemsize * pos_terms_occur.shape[1]))
        )
        sort_order = np.argsort(symp_matrix_view.ravel())
        s_index = sort_order[0]

        return s_index


    def _recursive_seq_rotations(self, AC_op: PauliwordOp) -> "PauliwordOp":
        """
        Args:
            AC_op (PauliwordOp): The AntiCommutingOp to apply sequence rotations to.

        Returns:
            PauliwordOp: The AntiCommutingOp after applying sequence rotations.
        """
        if AC_op.n_terms == 1:
            return AC_op
        else:
            # s_index fixed to zero (re-order done in unitary_partitioning method!)
            s_index = 0
            k_index = 1

            op_for_rotation = AC_op.copy()

            # take β_s P_s
            P_s = PauliwordOp(op_for_rotation.symp_matrix[s_index], [1])
            β_s = op_for_rotation.coeff_vec[s_index]
            β_k = op_for_rotation.coeff_vec[k_index]

            theta_sk = np.arctan(β_k / β_s)
            if β_s.real < 0:
                theta_sk = theta_sk + np.pi

            # check
            assert (np.isclose((β_k * np.cos(theta_sk) - β_s * np.sin(theta_sk)), 0)), 'term not zeroing out'

            # -X_sk = -1j * Ps @ Pk
            jP_k = PauliwordOp(op_for_rotation.symp_matrix[k_index], [-1j])
            X_sk = P_s * jP_k
            if X_sk.coeff_vec[0].real < 0:
                X_sk.coeff_vec[0] *= -1
                theta_sk *= -1

            self.X_sk_rotations.append((X_sk, theta_sk))

            # update coeffs
            op_for_rotation.coeff_vec[s_index] = np.sqrt(β_s ** 2 + β_k ** 2)
            op_for_rotation.coeff_vec[k_index] = 0

            # build op without k term and (included modified s term)
            AC_op_rotated = PauliwordOp(np.delete(op_for_rotation.symp_matrix, k_index, axis=0),
                                        np.delete(op_for_rotation.coeff_vec, k_index, axis=0))

            ## know how operator acts therefore don't need to actually do rotations

            return self._recursive_seq_rotations(AC_op_rotated)

    def unitary_partitioning(self, s_index: int=None, up_method: Optional[str]='seq_rot') \
            -> Tuple[PauliwordOp, Union[PauliwordOp, List[Tuple[PauliwordOp, float]]], float, "AntiCommutingOp"]:
        """
        Apply unitary partitioning on anticommuting operator (self)

        Args:
            s_index (int): index of row in symplectic matrix that defines Pauli operator to reduce too (Ps).
                           if set to None then code will find least dense Pauli operator and use that.
            up_method (str): unitary partitoning method ['LCU', 'seq_rot']

        Returns:
            Ps (PauliwordOp): Pauli operator of term reduced too
            rotations (PauliwordOp): rotations to perform unitary partitioning
            gamma_l (float): normalization constant of clique (anticommuting operator)
            AC_op (AntiCommutingOp): normalized clique - i.e. self == gamma_l * AC_op
        """
        assert up_method in ['LCU', 'seq_rot'], f'unknown unitary partitioning method: {up_method}'

        if s_index is None:
            s_index = self.get_least_dense_term_index()

        if np.isclose(self.coeff_vec[s_index], 0):
            # need to correct for s_index having zero coeff...
            s_index = np.argmax(abs(self.coeff_vec))
            warnings.warn(f's indexed term has zero coeff, s_index set to {s_index} so that nonzero operator is rotated onto')
       
        s_index = int(s_index)
        BsPs    = self[s_index]

        # NOTE: term to reduce to is at the top of sym matrix i.e. s_index of ZERO now!
        no_BsPs = (self - BsPs).cleanup()
        if (len(no_BsPs.coeff_vec)==1 and no_BsPs.coeff_vec[0]==0):
            AC_op = BsPs
        else:
            AC_op = BsPs.append(no_BsPs)

        if AC_op.n_terms == 1:
            rotations = []
            gamma_l = np.linalg.norm(AC_op.coeff_vec)
            AC_op.coeff_vec = AC_op.coeff_vec / gamma_l
            Ps = AC_op
            return Ps, rotations, gamma_l, self.multiply_by_constant(1/gamma_l)

        else:

            assert np.isclose(np.sum(AC_op.coeff_vec.imag), 0), 'cannot apply unitary partitioning to operator with complex coeffs'

            gamma_l = np.linalg.norm(AC_op.coeff_vec)
            AC_op.coeff_vec = AC_op.coeff_vec / gamma_l

            if up_method=='seq_rot':
                if len(self.X_sk_rotations)!=0:
                    self.X_sk_rotations = []
                Ps = self._recursive_seq_rotations(AC_op)
                rotations = self.X_sk_rotations
            elif up_method=='LCU':
                if self.R_LCU is not None:
                    self.R_LCU = None

                Ps = self.generate_LCU_operator(AC_op)
                rotations = LCU_as_seq_rot(self.R_LCU)
            else:
                raise ValueError(f'unknown unitary partitioning method: {up_method}!')

            return Ps, rotations, gamma_l, self.multiply_by_constant(1/gamma_l)
        
    def multiply_by_constant(self, constant: float) -> "AntiCommutingOp":
        """ Return AntiCommutingOp under constant multiplication
        """
        AC_op_copy = self.copy()
        AC_op_copy.coeff_vec *= constant
        return AC_op_copy
    
    @classmethod
    def random(cls, n_qubits: int, n_terms: Union[None, int]=None, apply_clifford=True) -> "AntiCommutingOp":
        """
        generate a random real coefficient anticommuting op

        """
        from symmer.utils import random_anitcomm_2n_1_PauliwordOp
        if n_terms is None:
            n_terms = 2*n_qubits+1

        assert n_terms<= 2*n_qubits+1, f'cannot have {n_terms} Pops on {n_qubits} qubits'
        return cls.from_PauliwordOp( random_anitcomm_2n_1_PauliwordOp(n_qubits, apply_clifford=apply_clifford)[:n_terms])

    def generate_LCU_operator(self, AC_op) -> PauliwordOp:
        """
        Given the normalized anticommuting operator object, function takes current symp ordering and
        finds the linear combination of unitaries (LCU) (https://arxiv.org/pdf/1908.08067.pdf) to reduce the operator
        to a single Pauli operator in the operator.

        Note if s_index is set to none then the least dense term is reduced too. Unlike the gen_seq_rotations method
        this approach doesn't acutally matter what order the objects symplectic matrix.


        Args:
            s_index (int): optional integar defining index of term to reduce too. Note if not set then defaults to
                           least dense Pauli operator in AC operator.
            check_reduction (bool): flag to check reduction by applying LCU on original operator.

        Returns:
            R_LCU (PauliwordOp): PauliwordOp that is a linear combination of unitaries
            P_s (PauliwordOp): single PauliwordOp that has been reduced too.
        """
        ## s_index is ensured to be in zero position in unitary_partitioning method! 
        ## if using function without this method need to ensure term to rotate onto is the zeroth index of AC_op!
        s_index=0

        # note gamma_l norm applied on init!
        Ps_LCU = PauliwordOp(AC_op.symp_matrix[s_index], [1])
        βs = AC_op.coeff_vec[s_index]

        #  ∑ β_k 𝑃_k  ... note this doesn't contain 𝛽_s 𝑃_s
        no_βsPs = AC_op - (Ps_LCU.multiply_by_constant(βs))

        # Ω_𝑙 ∑ 𝛿_k 𝑃_k  ... renormalized!
        omega_l = np.linalg.norm(no_βsPs.coeff_vec)
        no_βsPs.coeff_vec = no_βsPs.coeff_vec / omega_l

        phi_n_1 = np.arccos(βs)
        # require sin(𝜙_{𝑛−1}) to be positive...
        if (phi_n_1 > np.pi):
            phi_n_1 = 2 * np.pi - phi_n_1

        alpha = phi_n_1
        I_term = 'I' * Ps_LCU.n_qubits
        self.R_LCU = PauliwordOp.from_dictionary({I_term: np.cos(alpha / 2)})

        sin_term = -np.sin(alpha / 2)

        for dkPk in no_βsPs:
            dk_PkPs = dkPk * Ps_LCU
            self.R_LCU += dk_PkPs.multiply_by_constant(sin_term)

        return Ps_LCU

def LCU_as_seq_rot(R_LCU: PauliwordOp) -> List[Tuple[PauliwordOp, float]]:
    """
    Convert a unitary composed of a
    See equations 18 and 19 of https://arxiv.org/pdf/1907.09040.pdf

    number of rotations is 2*(R_LCU.n_terms-1), which can at most be 4*n_qubits

    Args:
        R_LCU (PauliwordOp): unitary composed as a normalized linear combination of imaginary anticommuting Pauli operators (excluding identity)
    Returns:
        expon_p_terms (list): list of rotations generated by Pauli operators to implement AC_op unitary

    ** Example use **

    from symmer.utils import random_anitcomm_2n_1_PauliwordOp
    from symmer.operators import AntiCommutingOp
    from symmer.evolution.exponentiation import exponentiate_single_Pop
    from symmer.utils import product_list

    nq = 3 # change (do not make too large as has exp checking cost)

    AC_2n1 = random_anitcomm_2n_1_PauliwordOp(nq)
    AC_op = AntiCommutingOp.from_PauliwordOp(AC_2n1)
    Ps_LCU, rotations_LCU, gamma_l, AC_normed = AC_op.unitary_partitioning(s_index=0, up_method= 'LCU')
    print(AC_normed.perform_rotations(rotations_LCU) == Ps_LCU)

    ## expensive check to see if operation is identical! should NOT do this when using
    a2 = product_list([exponentiate_single_Pop(P.multiply_by_constant(1j*angle/2)) for P, angle in rotations_LCU])
    print(AC_op.R_LCU == a2)
    """
    if isinstance(R_LCU, list) and len(R_LCU)==0:
        # case where there are no rotations
        return list()
    
    assert R_LCU.n_terms > 1, 'AC_op must have more than 1 term'
    assert np.isclose(np.linalg.norm(R_LCU.coeff_vec), 1), 'AC_op must be l2 normalized'

    expon_p_terms = []

    # # IF imaginary components the this makes real (but need phase correction later!)
    coeff_vec = R_LCU.coeff_vec.real + R_LCU.coeff_vec.imag

    # for k, c_k in enumerate(coeff_vec):
    #     P_k = R_LCU[k]
    #     theta_k = np.arcsin(c_k / np.linalg.norm(coeff_vec[:(k + 1)]))
    #     P_k.coeff_vec[0] = 1
    #     expon_p_terms.append(tuple((P_k, theta_k)))
    # ## phase correction - change angle by -pi in first rotation!
    # expon_p_terms[0] = (expon_p_terms[0][0], expon_p_terms[0][1]-np.pi)

    for k in range(1, R_LCU.n_terms):
        P_k = R_LCU[k]
        c_k = coeff_vec[k]
        theta_k = np.arcsin(c_k / np.linalg.norm(coeff_vec[:(k + 1)]))
        P_k.coeff_vec[0] = 1
        expon_p_terms.append(tuple((P_k, theta_k)))

    expon_p_terms = [*expon_p_terms, *expon_p_terms[::-1]]
    
    return expon_p_terms

# from symmer.operators.utils import mul_symplectic
# def conjugate_Pop_with_R(Pop:PauliwordOp,
#                         R: PauliwordOp) -> PauliwordOp:
#     """
#     For a defined linear combination of pauli operators : R = ∑_{𝑖} ci Pi ... (note each P self-adjoint!)
#
#     perform the adjoint rotation R op R† =  R [∑_{a} ca Pa] R†
#
#     Args:
#         R (PauliwordOp): operator to rotate Pop by
#     Returns:
#         rot_H (PauliwordOp): rotated operator
#
#     ### Notes
#     R = ∑_{𝑖} ci Pi
#     R^{†} = ∑_{j}  cj^{*} Pj
#     note i and j here run over the same indices!
#     apply R H R^{†} where H is Pop (current Pauli defined in class object)
#
#     ### derivation:
#
#     = (∑_{𝑖} ci Pi ) * (∑_{a} ca Pa ) * ∑_{j} cj^{*} Pj
#
#     = ∑_{a}∑_{i}∑_{j} (ci ca cj^{*}) Pi  Pa Pj
#
#     # can write as case for when i==j and i!=j
#
#     = ∑_{a}∑_{i=j} (ci ca ci^{*}) Pi  Pa Pi + ∑_{a}∑_{i}∑_{j!=i} (ci ca cj^{*}) Pi  Pa Pj
#
#     # let C by the termwise commutator matrix between H and R
#     = ∑_{a}∑_{i=j} (-1)^{C_{ia}} (ci ca ci^{*}) Pa  + ∑_{a}∑_{i}∑_{j!=i} (ci ca cj^{*}) Pi  Pa Pj
#
#     # next write final term over upper triange (as i and j run over same indices)
#     ## so add common terms for i and j and make j>i
#
#     = ∑_{a}∑_{i=j} (-1)^{C_{ia}} (ci ca ci^{*}) Pa
#       + ∑_{a}∑_{i}∑_{j>i} (ci ca cj^{*}) Pi  Pa Pj + (cj ca ci^{*}) Pj  Pa Pi
#
#     # then need to know commutation relation betwen terms in R
#     ## given by adjaceny matrix of R... here A
#
#
#     = ∑_{a}∑_{i=j} (-1)^{C_{ia}} (ci ca ci^{*}) Pa
#      + ∑_{a}∑_{i}∑_{j>i} (ci ca cj^{*}) Pi  Pa Pj + (-1)^{C_{ia}+A_{ij}+C_{ja}}(cj ca ci^{*}) Pi  Pa Pj
#
#
#     = ∑_{a}∑_{i=j} (-1)^{C_{ia}} (ci ca ci^{*}) Pa
#      + ∑_{a}∑_{i}∑_{j>i} (ci ca cj^{*} + (-1)^{C_{ia}+A_{ij}+C_{ja}}(cj ca ci^{*})) Pi  Pa Pj
#
#
#     """
#     if Pop.n_terms == 1:
#         rot_H = (R * Pop * R.dagger).cleanup()
#     else:
#         # anticommutes == 1 and commutes == 0
#         commutation_check = (~Pop.commutes_termwise(R)).astype(int)
#         adj_matrix = (~R.adjacency_matrix).astype(int)
#
#         c_list = R.coeff_vec
#         ca_list = Pop.coeff_vec
#
#         # rot_H = PauliwordOp.empty(Pop.n_qubits)
#         coeff_list = []
#         sym_vec_list = []
#         for ind_a, Pa_vec in enumerate(Pop.symp_matrix):
#             for ind_i, Pi_vec in enumerate(R.symp_matrix):
#                 sign = (-1) ** (commutation_check[ind_a, ind_i])
#
#                 coeff = c_list[ind_i] * c_list[ind_i].conj() * ca_list[ind_a]
#                 # rot_H += PauliwordOp(Pa_vec, [coeff * sign])
#                 sym_vec_list.append(Pa_vec)
#                 coeff_list.append(coeff * sign)
#
#                 # PiPa = PauliwordOp(Pi_vec, [1]) * PauliwordOp(Pa_vec, [1])
#                 phaseless_prod_PiPa, PiPa_coeff_vec = mul_symplectic(Pi_vec, 1,
#                                                                      Pa_vec, 1)
#
#                 for ind_j, Pj_vec in enumerate(R.symp_matrix[ind_i + 1:]):
#                     ind_j += ind_i + 1
#
#                     sign2 = (-1) ** (commutation_check[ind_a, ind_i] +
#                                      commutation_check[ind_a, ind_j] +
#                                      adj_matrix[ind_i, ind_j])
#
#                     coeff1 = c_list[ind_i] * ca_list[ind_a] * c_list[ind_j].conj()
#                     coeff2 = c_list[ind_j] * ca_list[ind_a] * c_list[ind_i].conj()
#
#                     overall_coeff = (coeff1 + sign2 * coeff2)
#                     if overall_coeff:
#                         # calculate PiPa term outside of j loop
#                         # overall_coeff_PiPaPj = PiPa * PauliwordOp(Pj_vec, [overall_coeff])
#                         # rot_H += overall_coeff_PiPaPj
#
#                         phaseless_prod, coeff_vec = mul_symplectic(phaseless_prod_PiPa, PiPa_coeff_vec,
#                                                                    Pj_vec, overall_coeff)
#                         sym_vec_list.append(phaseless_prod)
#                         coeff_list.append(coeff_vec)
#
#         rot_H = PauliwordOp(np.array(sym_vec_list),
#                             coeff_list).cleanup()
#     return rot_H
#



import os
import sys
import numpy as np
import quimb
from warnings import warn
from multiprocessing import Process, Queue, set_start_method

if sys.platform.lower() in ['linux', 'darwin', 'linux2']:
    set_start_method('fork', force = True)
else:
    set_start_method('spawn', force = True)

class ProcessHandler:

    if sys.platform.lower() in ['linux', 'darwin', 'linux2']:
        method  = 'mp'
    else:
        # method  = 'ray' # note ray no longer a dependency - must install manually to use
        method = 'single_thread'

    verbose = False

    def __init__(self):
        self.n_logical_cores = os.cpu_count()

    def prepare_chunks(self, iter):
        """ split a list into smaller sized chunks
        """
        iter = list(iter)
        self.n_chunks = min(len(iter), self.n_logical_cores)
        chunk_size = int(np.ceil(len(iter)/self.n_chunks))
        indices = np.append(np.arange(self.n_chunks)*chunk_size, None)
        for i,j in zip(indices[:-1], indices[1:]):
            yield iter[i:j]

    def _process_ray(self, func, iter, shared):
        """ Helper function for ray processing
        """
        warn("ray is being deprecated in symmer and is no longer a dependency - must install ray manually to use.")
        from ray import remote, put, get # deprecating, no longer a dependency

        if self.verbose:
            print(f'*** executing in ray mode ***')
        # duplicate func with ray.remote wrapper :
        @remote(num_cpus=self.n_logical_cores,
            runtime_env={
                "env_vars": {
                    "NUMBA_NUM_THREADS":   os.getenv("NUMBA_NUM_THREADS"),
                    "OMP_NUM_THREADS":     os.getenv("NUMBA_NUM_THREADS"),
                    "NUMEXPR_MAX_THREADS": str(self.n_logical_cores)
                }
            }
        )
        def _func(iter, shared):
            return func(iter, shared)
        # place into shared memory:
        shared_obj = put(shared)
        # split iterable into smaller chunks and parallelize remote instances:
        results = get(
            [
                _func.remote(chunk, shared_obj) 
                for chunk in self.prepare_chunks(iter)
            ]
        )
        # flatten the list and return:
        return [a for b in results for a in b]
        
    def _process_mp(self, func, iter, shared):
        """ Helper function for multiprocessing
        """
        if self.verbose:
            print(f'*** executing in multiprocessing mode ***')
        # wrapper function for putting results into queue
        def _func(iter, shared, _order=None, _queue=None):
            data_out = func(iter, shared)
            _queue.put((_order, data_out))

        chunks = list(self.prepare_chunks(iter))
        procs = [] # for storing processes
        queue = Queue(self.n_chunks) # storage of data from processes
        for index,chunk in enumerate(chunks): # index to ensure procs returned in correct order
            proc = Process(target=_func, args=(chunk, shared, index, queue))
            procs.append(proc)
            proc.start()
        # retrieve data from the queue
        data = []
        for _ in range(self.n_chunks):
            data.append(queue.get())
        # complete the processes
        for proc in procs:
            proc.join()
        # sort by correct chunk ordering and flatten data
        _,data = zip(*sorted(data))
        data = [a for b in data for a in b]
        return data
    
    def _process_single(self, func, iter, shared):
        """ Helper function for single threading
        """
        if self.verbose:
            print(f'*** executing in single-threaded mode ***')
        return func(iter, shared)
    
    def parallelize(self, func):

        def wrapper(iter, shared):

            _func = lambda iter,shared: [func(i, shared) for i in iter]

            if self.method   == 'mp':
                return self._process_mp(_func, iter, shared)
            elif self.method == 'ray':
                return self._process_ray(_func, iter, shared)
            elif self.method == 'single_thread':
                return self._process_single(_func, iter, shared)
            else:
                raise ValueError(f'Invalid processing method {self.method}, must be ray, mp or single_thread.')
            
        return wrapper
    
process = ProcessHandler()
    
if __name__ == '__main__':
            
    @process.parallelize
    def multiply_list(iter, shared):
        return [i*shared for i in iter]
    
    l = list(range(100))

    process.method = 'single_thread'
    print(multiply_list(l,2))
    process.method = 'mp'
    print(multiply_list(l,2))
    process.method = 'ray'
    print(multiply_list(l,2))




import numpy as np
from ncon import ncon
from typing import Union, List, Dict
from symmer.operators import PauliwordOp, QuantumState
from copy import copy
from functools import cached_property
from quimb.tensor.tensor_1d import MatrixProductOperator, MatrixProductState
from quimb.tensor.tensor_dmrg import DMRG2


class MPOOp:
    """
    Class to build MPO operator from Pauli strings and coeffs.
    """

    def __init__(self,
            pauliList: List[str],
            coeffList: List[complex],
            Dmax: int = None) -> None:
        """
        Initialize an MPO to represent an operator from Pauli strings and
        coefficients. MPO tensors are shapes as (σ, l, i, j) where σ is the
        physical leg, l is the output leg and i and j are the remaining
        legs.

        Args:
            pauliList (List[str]): List of Pauli Strings
            coeffList (List[complex]): List of complex coefficients
            Dmax (int): Maximum bond dimension. By default it is set to 'None'.
        """
        self.mpo = pstrings_to_mpo_optimized(pauliList, coeffList, Dmax)

    @classmethod
    def from_dictionary(cls,
            operator_dict: Dict[str, complex],
            Dmax: int = None) -> "MPOApproximator":
        """
        Initalize MPOApproximator using Pauli terms and coefficients stored in
        a dictionary like {pauli: coeff}

        Args:
            operator_dict (Dict[str, complex]): Dictionary of Pauli string terms and their coefficients.
            Dmax (int): Maximum bond dimension. By default it is set to 'None'.

        Returns:
            Matrix Product Operator (MPO) Approximator
        """
        paulis, coeffs = zip(*operator_dict.items())
        paulis = list(paulis)
        return cls(paulis, coeffs, Dmax)

    @classmethod
    def from_WordOp(cls,
            WordOp: PauliwordOp) -> "MPOApproximator":
        """
        Initialize MPOApproximator using PauliwordOp.

        Args:
            WordOp (PauliwordOp): PauliwordOp to initialize MPO Approximator.

        Returns:
            Matrix Product Operator (MPO) Approximator
        """
        return cls.from_dictionary(WordOp.to_dictionary())

    @cached_property
    def to_matrix(self) -> np.ndarray:
        '''
        Contract MPO to produce matrix representation of operator.

        Returns:
            Matrix Representation(np.ndarray) of operator.
        '''
        mpo = self.mpo
        contr = mpo[0]
        for tensor in mpo[1:]:
            σ1, l1, i1, j1 = contr.shape
            σ2, l2, i2, j2 = tensor.shape
            contr = ncon([contr, tensor], ((-1, -3, -5, 1), (-2, -4, 1, -6)))
            contr = np.reshape(contr, (σ1 * σ2, l1 * l2, i1, j2))

        contr = np.squeeze(contr)
        return contr
    
def get_MPO(operator: PauliwordOp, max_bond_dimension: int) -> MPOOp:
    """ 
    Return the Matrix Product Operator (MPO) of a PauliwordOp 
    (linear combination of Paulis) given a maximum bond dimension.

    Args: 
        operator (PauliwordOp): PauliwordOp (linear combination of Paulis)
        max_bond_dimension (int): Maximum bond dimension.

    Returns:
        The Matrix Product Operator (MPO) of a PauliwordOp passed as parameter.
    """
    pstrings, coefflist = zip(*operator.to_dictionary.items())
    mpo = MPOOp(pstrings, coefflist, Dmax=max_bond_dimension)
    return mpo

def find_groundstate_quimb(MPOOp: MPOOp, dmrg=None, gs_guess=None) -> QuantumState:
    """
    Use quimb's DMRG2 optimiser to approximate groundstate of MPOOp

    Args:
        MPOOp: MPOOp representing operator.
        dmrg: Quimb DMRG solver class. By default it is set to 'None'.
        gs_guess: Guess for the ground state, used as intialisation for the DMRG optimiser. Represented as a dense array. By default it is set to 'None'.
    Returns:
        dmrg_state (QuantumState): Approximated groundstate.
    """
    mpo = [np.squeeze(m) for m in MPOOp.mpo]
    MPO = MatrixProductOperator(mpo, 'dulr')

    if gs_guess is not None:
        no_qubits = int(np.log2(gs_guess.shape[0]))
        dims = [2] * no_qubits
        gs_guess = MatrixProductState.from_dense(gs_guess, dims)

    # Useful default for DMRG optimiser
    if dmrg is None:
        dmrg = DMRG2(MPO, bond_dims=[10, 20, 100, 100, 200], cutoffs=1e-10, p0=gs_guess)
    dmrg.solve(verbosity=0, tol=1e-6)

    dmrg_state = dmrg.state.to_dense()
    dmrg_state = QuantumState.from_array(dmrg_state).cleanup(zero_threshold=1e-5)

    return dmrg_state


Paulis = {
        'I': np.eye(2, dtype=np.complex64),
        'X': np.array([[0, 1],
                       [1, 0]], dtype=np.complex64),
        'Y': np.array([[0, -1j],
                       [1j, 0]], dtype=np.complex64),
        'Z': np.array([[1, 0],
                       [0, -1]], dtype=np.complex64),
        }

def coefflist_to_complex(coefflist):
    '''
    Convert a list of real + imaginary components into a complex vector.

    Args:
        coeffList (List): List of complex coefficients.

    Returns:
        Array of complex vectors
    '''
    arr = np.array(coefflist, dtype=complex)

    return arr[:, 0] + 1j*arr[:, 1]

def pstrings_to_mpo_optimized(pstrings, coeffs=None, Dmax=None):
    ''' 
    Convert a list of Pauli Strings into an MPO. If coeff list is given,
    rescale each Pauli string by the corresponding element of the coeff list.
    Bond dim specifies the maximum bond dimension, if None, no maximum bond
    dimension.
    Optimization is achieved by constructing final sum terms directly from:
        1) First letters of each Pauli string for the first term 
        2) Last letters of each Pauli string for the last term 
        3) Diagonal matrix per each part of 4D tensor with respective middle letters of Pauli strings 

    Args:
        pstrings (List[str]): List of Pauli Strings
        coeffs (List[complex]): List of coefficients. By default it is set to 'None'. If coeff list is given, each Pauli string is rescaled by the corresponding element of the coeff list.
        Dmax (int): Maximum bond dimension. By default it is set to 'None'.

    Returns:
        mpo: The Matrix Product Operator (MPO)
    '''

    if coeffs is None:
        coeffs = np.ones(len(pstrings))

    if Dmax is None:
        Dmax = np.inf

    summed = [None] * len(pstrings[0])
    first_p = np.array([Paulis[pstr[0]] for pstr in pstrings])

    for i in range(len(coeffs)):
        first_p[i] = np.multiply(first_p[i], coeffs[i])

    last_p = [pstr[-1] for pstr in pstrings]
    p_i = [(0, 0), (0, 1), (1, 0), (1, 1)]

    first_sum = [ [ [[ ]], [[ ]] ], [ [[ ]], [[ ]] ]]
    for p in first_p:
        for x, y in p_i:
            first_sum[x][y][0].extend([p[x][y]])

    summed[0] = np.array(first_sum)
    

    last_sum = [ [ [ ], [ ] ], [ [ ], [ ] ]]
    for p in last_p:
        for x, y in p_i:
            last_sum[x][y].append([ Paulis[p][x][y]])
   
    summed[-1] = np.array(last_sum)

    
    for i in range(1, len(pstrings[0]) - 1):
        middle_p_i = [pstr[i] for pstr in pstrings]
        new_tensor = [ [ [ ], [ ] ], [ [ ], [ ] ] ]

        for x, y in p_i:
            new_tensor[x][y] = np.diag(np.array([ Paulis[p][x][y] for p in middle_p_i ]))
        summed[i] = np.array(new_tensor)

    mpo = truncate_MPO(summed, Dmax)
    return mpo

def pstrings_to_mpo(pstrings, coeffs=None, Dmax=None):
    ''' 
    Convert a list of Pauli Strings into an MPO. If coeff list is given,
    rescale each Pauli string by the corresponding element of the coeff list.
    Bond dim specifies the maximum bond dimension, if None, no maximum bond
    dimension.

    Args:
        pstrings (List[str]): List of Pauli Strings
        coeffs (List[complex]): List of coefficients. By default it is set to 'None'. If coeff list is given, each Pauli string is rescaled by the corresponding element of the coeff list.
        Dmax (int): Maximum bond dimension. By default it is set to 'None'.

    Returns:
        mpo: The Matrix Product Operator (MPO)
    '''
    if coeffs is None:
        coeffs = np.ones(len(pstrings))

    if Dmax is None:
        Dmax = np.inf

    mpo = pstring_to_mpo(pstrings[0], coeffs[0])

    for pstr, coeff in zip(pstrings[1:], coeffs[1:]):
        _mpo = pstring_to_mpo(pstr, coeff)
        mpo = sum_mpo(mpo, _mpo)
        mpo = truncate_MPO(mpo, Dmax)

    return mpo

def pstring_to_mpo(pstring, scaling=None):
    '''
    Args:
        pstring (str): Pauli String
        scaling (complex): Scale the Pauli string by a factor of 'scaling'. By default it is set to 'None'.
    
    Returns:
        The Matrix Product Operator (MPO)
    '''
    As = []
    for p in pstring:
        pauli = Paulis[p]
        pauli_tensor = np.expand_dims(pauli, axis=(2, 3))
        As.append(pauli_tensor)

    if scaling is not None:
        As[0] = As[0] * scaling
    return As


def truncated_SVD(M, Dmax=None):
    '''
    Args:
        M: (..., P, Q) array_like. A real or complex array with M.ndim >= 2.
        Dmax (int): Maximum bond dimension. By default it is set to 'None'.

    Returns:
        U, S, V: Singular Value Decomposition of M
    '''
    U, S, V = np.linalg.svd(M, full_matrices=False)

    if Dmax is not None and len(S) > Dmax:
        S = S[:Dmax]
        U = U[:, :Dmax]
        V = V[:Dmax, :]

    return U, S, V

def truncate_MPO(mpo, Dmax):
    '''
    Args:
        mpo: Matrix Product Operator (MPO)
        Dmax (int): Maximum bond dimension.

    Returns:
        As: Truncated Matrix Product Operator.
    '''
    As = []
    for n in range(len(mpo) - 1):  # Don't need to run on the last term
        A = mpo[n]
        σ, l, i, j = A.shape
        A = A.reshape(σ * l * i, j)
        U, S, V = truncated_SVD(A, Dmax)
        D = len(S)

        # Update current term
        _A = U.reshape(σ, l, i, D)
        As.append(_A)

        # Update the next term
        M = np.diag(S) @ V
        _A1 = ncon([M, mpo[n+1]], ((-3, 1), (-1, -2, 1, -4)))
        mpo[n+1] = _A1

    As.append(mpo[-1])

    return As

def sum_mpo(mpo1, mpo2):
    '''
    Args:
        mpo1: First Matrix Product Operator (MPO)
        mpo2: Second Matrix Product Operator (MPO)
    
    Returns:
        summed: Sum of First and Second Matrix Product Operator.
    '''
    summed = [None] * len(mpo1)
    σ10, l10, i10, j10 = mpo1[0].shape
    σ20, l20, i20, j20 = mpo2[0].shape
    t10 = copy(mpo1[0])
    t20 = copy(mpo2[0])
    first_sum = np.zeros((σ10, l10, i10, j10+j20), dtype=complex)
    first_sum[:, :, :, :j10] = t10
    first_sum[:, :, :, j10:] = t20
    summed[0] = first_sum

    σ1l, l1l, i1l, j1l = mpo1[-1].shape
    σ2l, l2l, i2l, j2l = mpo2[-1].shape
    t1l = copy(mpo1[-1])
    t2l = copy(mpo2[-1])
    first_sum = np.zeros((σ1l, l1l, i1l + i2l, j1l), dtype=complex)
    first_sum[:, :, :i1l, :] = t1l
    first_sum[:, :, i1l:, :] = t2l
    summed[-1] = first_sum
    for i in range(1, len(mpo1) - 1):
        σ1, l1, i1, j1 = mpo1[i].shape
        σ2, l2, i2, j2 = mpo2[i].shape
        t1 = copy(mpo1[i])
        t2 = copy(mpo2[i])


        new_shape = (σ1, l1, i1+i2, j1+j2)

        new_tensor = np.zeros(new_shape, dtype=complex)

        new_tensor[:, :, :i1, :j1] = t1
        new_tensor[:, :, i1:, j1:] = t2
        summed[i] = new_tensor

    return summed


import numpy as np
from typing import List
from functools import reduce
from symmer.operators import PauliwordOp

def exponentiate_single_Pop(P: PauliwordOp) -> PauliwordOp:
    """
    Exponentiate a single Pauli term as e^{P}

    If goal is to implement e^{iθP} then coefficient of P must be iθ (note imaginary part must be included in coeff)

    Args:
        P (PauliwordOp): Pauli operator to exponentiate
    Returns:
        exp_P (PauliwordOp): PauliwordOp representation of exponentiated operator
    """
    assert(P.n_terms == 1), 'Can only exponentiate single Pauli terms'
    P_copy = P.copy()
    P_copy.coeff_vec[0] = 1
    exp_P = (
        (P_copy**0).multiply_by_constant(np.cosh(P.coeff_vec[0])) +
        (P_copy**1).multiply_by_constant(np.sinh(P.coeff_vec[0]))
    )
    return exp_P

def trotter(op:PauliwordOp, trotnum:int=1) -> PauliwordOp:
    """ 
    Computes the exponential exp(op). This is exact only when 
    op is fully commuting, otherwise approximates the exponential
    and increasing trotnum will improve precision.

    Args:
        op (PauliwordOp): Pauli operator to exponentiate
        tortnum (int): Increasing trotnum will improve precision when exact exponential is not computed. By default, it is set to 1.
    """
    op_copy = op.copy().multiply_by_constant(1/trotnum)
    factors = [exponentiate_single_Pop(P) for P in op_copy]*trotnum
    return reduce(lambda x,y:x*y, factors)

def truncated_exponential(op:PauliwordOp, truncate_at:int=10) -> PauliwordOp:
    raise NotImplementedError


from symmer.operators import PauliwordOp, QuantumState, AntiCommutingOp
import numpy as np
import scipy as sp
from typing import List, Tuple, Union
from functools import reduce
from scipy.sparse import csr_matrix

def exact_gs_energy(
        sparse_matrix, 
        initial_guess=None, 
        n_particles=None, 
        number_operator=None, 
        n_eigs=6
    ) -> Tuple[float, np.array]:
    """ 
    Return the ground state energy and corresponding ground statevector for the input operator
    
    Specifying a particle number will restrict to eigenvectors |ψ> such that <ψ|N_op|ψ> = n_particles
    where N_op is the given number operator.

    Args:
        sparse_matrix (csr_matrix): The sparse matrix for which we want to compute the eigenvalues and eigenvectors.
        initial_guess (array): The initial guess for the eigenvectors.
        n_particles (int):  Particle number to restrict eigenvectors |ψ> such that <ψ|N_op|ψ> = n_particles where N_op is the given number operator.
        number_operator (array): Number Operator to restrict eigenvectors |ψ> such that <ψ|N_op|ψ> = n_particles.
        n_eigs (int): The number of eigenvalues and eigenvectors to compute.

    Returns:
        evl(float): The ground state energy for the input operator
        QState(QuantumState): Ground statevector for the input operator corresponding to evl.
    """
    if number_operator is None:
        # if no number operator then need not compute any further eigenvalues
        n_eigs = 1

    # Note the eigenvectors are stored column-wise so need to transpose
    if sparse_matrix.shape[0] > 2**5:
        eigvals, eigvecs = sp.sparse.linalg.eigsh(
            sparse_matrix,k=n_eigs,v0=initial_guess,which='SA',maxiter=1e7
        )
    else:
        # for small matrices the dense representation can be more efficient than sparse!
        eigvals, eigvecs = np.linalg.eigh(sparse_matrix.toarray())
    
    # order the eigenvalues by increasing size
    order = np.argsort(eigvals)
    eigvals, eigvecs = eigvals[order], eigvecs[:, order]
    
    if n_particles is None:
        # if no particle number is specified then return the smallest eigenvalue
        return eigvals[0], QuantumState.from_array(eigvecs[:,0].reshape([-1,1]))
    else:
        assert(number_operator is not None), 'Must specify the number operator.'
        # otherwise, search through the first n_eig eigenvalues and check the Hamming weight
        # of the the corresponding eigenvector - return the first match with n_particles
        for evl, evc in zip(eigvals, eigvecs.T):
            psi = QuantumState.from_array(evc.reshape([-1,1])).cleanup(zero_threshold=1e-5)
            assert(~np.any(number_operator.X_block)), 'Number operator not diagonal'
            expval_n_particle = 0
            for Z_symp, Z_coeff in zip(number_operator.Z_block, number_operator.coeff_vec):
                sign = (-1) ** np.einsum('ij->i', 
                    np.bitwise_and(
                        Z_symp, psi.state_matrix
                    )
                )
                expval_n_particle += Z_coeff * np.sum(sign * np.square(abs(psi.state_op.coeff_vec)))
            if np.round(expval_n_particle) == n_particles:
                return evl, QuantumState.from_array(evc.reshape([-1,1]))
        # if a solution is not found within the first n_eig eigenvalues then error
        raise RuntimeError('No eigenvector of the correct particle number was identified - try increasing n_eigs.')

def get_entanglement_entropy(psi: QuantumState, qubits: List[int]) -> float:
    """
    Get the Von Neumann entropy of the biprtition defined by the specified subsystem 
    qubit indices and those remaining (i.e. those that will be subsequently traced out)

    Args:
        psi (QuantumState): the quantum state for which we wish to extract the entanglement entropy
        qubits (List[int]): the qubit indices to project onto (the remaining qubits will be traced over)
    
    Returns:
        entropy (float): the Von Neumann entropy of the reduced subsystem
    """
    reduced = psi.get_rdm(qubits)
    eigvals, eigvecs = np.linalg.eig(reduced)
    eigvals = eigvals[eigvals>0]
    entropy = -np.sum(eigvals*np.log(eigvals)).real
    return entropy

def random_anitcomm_2n_1_PauliwordOp(n_qubits, complex_coeff=False, apply_clifford=True):
    """ 
    Generate a anticommuting PauliOperator of size 2n+1 on n qubits (max possible size)
    with normally distributed coefficients. Generates in structured way then uses Clifford rotation (default)
    to try and make more random (can stop this to allow FAST build, but inherenet structure
    will be present as operator is formed in specific way!)

    Args:
        n_qubits (int): Number of Qubits
        complex_coeff (bool): Boolean representing whether if we want complex coefficents or not.
        apply_clifford (bool): Boolean representing whether we want to apply clifford rotations to get rid of structure or not.

    Returns:
        P_anticomm (csr_matrix): Anticommuting PauliOperator of size 2n+1 on n qubits with normally distributed coefficients.
    """
    # base = 'X' * n_qubits
    # I_term = 'I' * n_qubits
    # P_list = [base]
    # for i in range(n_qubits):
    #     # Z_term
    #     P_list.append(base[:i] + 'Z' + I_term[i + 1:])
    #     # Y_term
    #     P_list.append(base[:i] + 'Y' + I_term[i + 1:])
    # coeff_vec = np.random.randn(len(P_list)).astype(complex)
    # if complex_coeff:
    #     coeff_vec += 1j * np.random.randn((len(P_list)))
    # P_anticomm = PauliwordOp.from_dictionary((dict(zip(P_list, coeff_vec))))

    Y_base = np.hstack((np.eye(n_qubits), np.tril(np.ones(n_qubits))))
    X_base = Y_base.copy()
    X_base[:, n_qubits:] = np.tril(np.ones(n_qubits), -1)

    ac_symp = np.vstack((Y_base, X_base))

    Z_symp = np.zeros(2 * n_qubits)
    Z_symp[n_qubits:] = np.ones(n_qubits)

    ac_symp = np.vstack((ac_symp, Z_symp)).astype(bool)

    coeff_vec = np.random.randn(ac_symp.shape[0]).astype(complex)
    if complex_coeff:
        coeff_vec += 1j * np.random.randn(2 * n_qubits + 1).astype(complex)

    P_anticomm = PauliwordOp(ac_symp, coeff_vec)

    if apply_clifford:
        # apply clifford rotations to get rid of structure
        U_cliff_rotations = []
        for _ in range(n_qubits * 5):
            P_rand = PauliwordOp.random(n_qubits, n_terms=1)
            P_rand.coeff_vec = [1]
            U_cliff_rotations.append((P_rand, np.random.choice([np.pi/2, -np.pi/2])))

        P_anticomm = P_anticomm.perform_rotations(U_cliff_rotations)

    assert P_anticomm.n_terms == 2 * n_qubits + 1

    ## expensive check
    # anti_comm_check = P_anticomm.adjacency_matrix.astype(int) - np.eye(P_anticomm.adjacency_matrix.shape[0])
    # assert(np.sum(anti_comm_check) == 0), 'operator needs to be made of anti-commuting Pauli operators'

    return P_anticomm


def tensor_list(factor_list:List[PauliwordOp]) -> PauliwordOp:
    """ 
    Given a list of PauliwordOps, recursively tensor from the right
    
    Args:
        factor_list (list): list of PauliwordOps
    
    Returns: 
        Tensor Product of items in factor_list from the right 
    """
    return reduce(lambda x,y:x.tensor(y), factor_list)


def product_list(product_list:List[PauliwordOp]) -> PauliwordOp:
    """ 
    Given a list of PauliwordOps, recursively take product from the right

    Args:
        product_list (list): list of PauliwordOps

    Returns:
        Product of items in product_list from the right 
    """
    return reduce(lambda x,y:x*y, product_list)


def gram_schmidt_from_quantum_state(state:Union[np.array, list, QuantumState]) ->np.array:
    """
    build a unitary to build a quantum state from the zero state (aka state defines first column of unitary)
    uses gram schmidt to find other (orthogonal) columns of matrix

    Args:
        state (np.array): 1D array of quantum state (size 2^N qubits)
    Returns:
        M (np.array): unitary matrix preparing input state from zero state
    """

    if isinstance(state, QuantumState):
        N_qubits = state.n_qubits
        state = state.to_sparse_matrix.toarray().reshape([-1])
    else:
        state = np.asarray(state).reshape([-1])
        N_qubits = round(np.log2(state.shape[0]))
        missing_amps = 2**N_qubits - state.shape[0]
        state = np.hstack((state, np.zeros(missing_amps, dtype=complex)))

    assert state.shape[0] == 2**N_qubits, 'state is not defined on power of two'
    assert np.isclose(np.linalg.norm(state), 1), 'state is not normalized'

    M = np.eye(2**N_qubits, dtype=complex)

    # reorder if state has 0 amp on zero index
    if np.isclose(state[0], 0):
        max_amp_ind = np.argmax(state)
        M[:, [0, max_amp_ind]] = M[:, [max_amp_ind,0]]

    # defines first column
    M[:, 0] = state
    for a in range(M.shape[0]):
        for b in range(a):
            M[:, a]-= (M[:, b].conj().T @ M[:, a]) * M[:, b]

        # normalize
        M[:, a] = M[:, a] / np.linalg.norm( M[:, a])

    return M


def matrix_allclose(A: Union[csr_matrix, np.array], B:Union[csr_matrix, np.array], tol:int = 1e-15) -> bool:
    """
    check matrix A and B have the same entries up to a given tolerance
    Args:
        A : matrix A
        B:  matrix B
        tol: allowed difference

    Returns:
        bool

    """
    if isinstance(A, csr_matrix) and isinstance(B, csr_matrix):
        max_diff = np.abs(A-B).max()
        return max_diff <= tol
    else:
        if isinstance(A, csr_matrix):
            A = A.toarray()

        if isinstance(B, csr_matrix):
            B = B.toarray()

        return np.allclose(A, B, atol=tol)


def get_PauliwordOp_root(power: int, pauli: PauliwordOp) -> PauliwordOp:
    """
    Get arbitrary power of a single Pauli operator. See eq1 in https://arxiv.org/pdf/2012.01667.pdf

    Log(A) in paper given by = 1j*pi*(I-P)/2 here

    P^{k} = e^{k i pi Q}

    Q = (I-P)/2, where P in {X,Y,Z}

    e^{k i pi (I-P)/2} = e^{k i pi/2 I} * e^{ - k i pi/2 P} <- expand product!

    Args:
        power (int): power to take
        pauli (PauliwordOp): Pauli operator to take power of
    Returns:
        Pk (PauliwordOp): Pauli operator that is power of input

    """
    assert pauli.n_terms == 1, 'can only take power of single operators'

    I_term = PauliwordOp.from_list(['I' * pauli.n_qubits])

    cos_term = np.cos(power * np.pi / 2)
    sin_term = np.sin(power * np.pi / 2)

    Pk = (I_term.multiply_by_constant(cos_term ** 2 + 1j * cos_term * sin_term) +
          pauli.multiply_by_constant(-1j * cos_term * sin_term + sin_term ** 2))

    return Pk


def Get_AC_root(power: float, operator: AntiCommutingOp) -> PauliwordOp:
    """
    Get arbitrary power of an anticommuting Pauli operator.

    ** test **
    from symmer.operators import AntiCommutingOp
    from symmer.utils import random_anitcomm_2n_1_PauliwordOp, Get_AC_root

    op = random_anitcomm_2n_1_PauliwordOp(3)
    AC = AntiCommutingOp.from_PauliwordOp(op)

    p = 0.25
    root = Get_AC_root(p, AC)
    print((root*root*root*root - AC).cleanup(zero_threshold=1e-12)

    Args:
        power (float): any power
        operator (AntiCommutingOp) Anticommuting Pauli operator

    Returns:
        AC_root (PauliwordOp): operator representing power of AC input

    """
    Ps, rot, gamma_l, AC_normed = operator.unitary_partitioning(up_method='LCU')

    Ps_root = get_PauliwordOp_root(power, Ps)

    AC_root = (rot.dagger * Ps_root * rot).multiply_by_constant(gamma_l ** power)

    return AC_root


import numpy as np
from itertools import combinations
from networkx import Graph, draw_spring
from networkx.algorithms.isomorphism.isomorphvf2 import GraphMatcher
from symmer import PauliwordOp
from collections import Counter
from typing import Union
from qiskit import QuantumCircuit

def get_CNOT_connectivity_graph(evolution_obj:Union[PauliwordOp,QuantumCircuit], print_graph=False):
    """ 
    Get the graph whoss edges denote nonlocal interaction between two qubits.
    This is useful for device-aware ansatz construction to ensure the circuit connectiviy
    may be accomodated by the topology of the target quantum processor. 

    Args:
        evolution_obj (Union[PauliwordOp, QuantumCircuit]): Evolution Object
        print_graph (bool): If True, the graph is drawn. By default, it's set to False.
    """
    if isinstance(evolution_obj, QuantumCircuit):
        edges = [[q.index for q in step[1]] for step in evolution_obj.data if step[0].name!='barrier' and len(step[1])>1]
        weighted_edges = [(u,v,w) for (u,v),w in Counter(edges).items()]
    else:
        rows, cols = np.where(evolution_obj.X_block | evolution_obj.Z_block)
        support_indices = [evolution_obj.n_qubits - 1 - cols[rows==i] for i in np.unique(rows)]
        qubit_coupling = [list(zip(x[:-1], x[1:])) for x in support_indices]
        edges = [a for b in qubit_coupling for a in b]
        weighted_edges = [(u,v,w*2) for (u,v),w in Counter(edges).items()]
        
    G = Graph()
    G.add_weighted_edges_from(weighted_edges)
    if print_graph:
        draw_spring(G)
    return G

def _subgraph_isomorphism_distance(G, target, depth=0):
    """
    """
    if depth == 0:
        if GraphMatcher(target, G).subgraph_is_isomorphic():
            return 0
        else:
            return None
    else:
        ordered_nodes = sorted(
            combinations(G.nodes, r=depth), 
            key=lambda nodes:-np.sum([len(G.edges(n)) for n in nodes])
        )
        for nodes in ordered_nodes:
            G_temp = G.copy()
            for n in nodes:
                G_temp.remove_node(n)
            if GraphMatcher(target, G_temp).subgraph_is_isomorphic():
                dropped_edge_weights = [G.edges[e]['weight'] for n in nodes for e in G.edges(n)]
                return sum(dropped_edge_weights)

        return None

def subgraph_isomorphism_distance(G, target, max_depth=3):
    """
    """
    depth = 0
    for depth in range(max_depth):
        dist = _subgraph_isomorphism_distance(G, target, depth)
        if dist is not None:
            return dist
        else:
            depth += 1
    return None

def topology_match_score(ansatz_operator, topology, max_depth=3):
    """
    """
    n_entangling_gates = 2*(np.count_nonzero(ansatz_operator.X_block | ansatz_operator.Z_block)-ansatz_operator.n_terms)
    if n_entangling_gates == 0:
        return 1
    else:
        entangling_graph = get_CNOT_connectivity_graph(ansatz_operator)
        subgraph_cost = subgraph_isomorphism_distance(entangling_graph, topology, max_depth=max_depth)
        if subgraph_cost is None:
            return 0
        else:
            return 1-subgraph_cost/n_entangling_gates




"""init for projection."""
from .utils import *
from .base import S3Projection
from .qubit_tapering import QubitTapering
from .contextual_subspace import ContextualSubspace
from .qubit_subspace_manager import QubitSubspaceManager


import numpy as np
import warnings
from functools import reduce
from typing import Dict, List, Tuple, Union
from symmer import process
from symmer.operators.utils import _rref_binary, _cref_binary, check_independent
from symmer.operators import PauliwordOp, QuantumState, symplectic_to_string, single_term_expval

class IndependentOp(PauliwordOp):
    """ 
    Special case of PauliwordOp, in which the operator terms must
    by algebraically independent, with all coefficients set to integers +/-1.

    - stabilizer_rotations
        This method determines a sequence of Clifford rotations mapping the
        provided stabilizers onto single-qubit Paulis (sqp), either X or Z

    <!> Note the target_sqp must be chosen BEFORE generating 
        the stabilizer rotations, since these will be cached.
    """
    def __init__(self,
            symp_matrix:   np.array,
            coeff_vec:  Union[List[complex], np.array] = None,
            target_sqp: str = 'Z'):
        """
        Args:
            symp_matrix (np.array): Symplectic matrix.
            coeff_vec (np.array): Coefficient Vector. By default, it's set to 'None'.
            target_sqp (str): Target SQP (Single-Qubit Pauli). By default it is to 'Z'.
        """
        if coeff_vec is None:
            coeff_vec = np.ones(symp_matrix.shape[0], dtype=complex)
        super().__init__(symp_matrix, coeff_vec)
        self._check_stab()
        self.coeff_vec = self.coeff_vec.real.astype(int)
        self._check_independent()
        if target_sqp in ['X', 'Z', 'Y']:
            self.target_sqp = target_sqp
        else:
            raise ValueError('Target single-qubit Pauli not recognised - must be X or Z')
        # set up these attributes to later track rotations mapping stabilizers to single-qubit Pauli operators
        self.stabilizer_rotations = None
        self.used_indices = None

    @classmethod
    def from_PauliwordOp(cls,
            PwordOp: PauliwordOp
        ) -> "IndependentOp":
        """
        Args:
            PwordOp (PauliwordOp): Pauli operator to be used to form the Independent Operator.

        Return:
            Independent Operator
        """
        return cls(PwordOp.symp_matrix, PwordOp.coeff_vec)

    @classmethod
    def from_list(cls, 
            pauli_terms :List[str], 
            coeff_vec:   List[complex] = None
        ) -> "IndependentOp":
        """
        Args:
            pauli_terms (List[str]): List of Pauli Terms.
            coeff_vec (List[complex]): List of complex coeffecients.

        Return:
            Independent Operator
        """
        PwordOp = super().from_list(pauli_terms, coeff_vec)
        return cls.from_PauliwordOp(PwordOp)

    @classmethod
    def from_dictionary(cls,
            operator_dict: Dict[str, complex]
        ) -> "IndependentOp":
        """ 
        Initialize a PauliwordOp from its dictionary representation {pauli:coeff, ...}

        Args:
            operator_dict (Dict[str, complex]): Dictionary representation of IndependentOp
                    
        Return:
            Independent Operator
        """
        PwordOp = super().from_dictionary(operator_dict)
        return cls.from_PauliwordOp(PwordOp)

    @classmethod
    def symmetry_generators(cls, 
            PwordOp: PauliwordOp, 
            commuting_override:bool=False,
            largest_clique = False
        ) -> "IndependentOp":
        """ 
        Identify a symmetry basis for the supplied Pauli operator with
        symplectic representation  M = [ X | Z ]. We perform columnwise 
        Gaussian elimination to yield the matrix

                [ Z | X ]     [ R ]
                |-------| ->  |---|
                [   I   ]     [ Q ]

        Indexing the zero columns of R with i, we form the matrix
        
                S^T = [ Q_i1 | ... | Q_iM ] 
                
        and conclude that S is the symplectic representation of the symmetry basis.
        This holds since MΩS^T=0 by construction, which implies commutativity.

        Since we only need to reduce columns, the algorithm scales with the number of
        qubits N, not the number of terms M, and is therefore at worst O(N^2).

        Args:
            PwordOp (PauliwordOp): Pauli operator.
            commuting_override (bool): If 'True', the check for commuting symmetries is overriden. By default, it's set to 'False'.
            largest_clique (bool): If 'True', the largest commuting subset is determined using clique cover or a greedy method. By default, it's set to 'False'.

        Returns:
            Symplectic representation of the symmetry basis.
        """
        # swap order of XZ blocks in symplectic matrix to ZX
        to_reduce = np.vstack([np.hstack([PwordOp.Z_block, PwordOp.X_block]), np.eye(2*PwordOp.n_qubits, dtype=bool)])
        cref_matrix = _cref_binary(to_reduce)
        S_symp = cref_matrix[PwordOp.n_terms:,np.all(~cref_matrix[:PwordOp.n_terms], axis=0)].T
        S = cls(S_symp, np.ones(S_symp.shape[0]))
        if S.n_terms==0:
            warnings.warn('The input PauliwordOp has no Z2 symmetries.')
            return S
            # raise RuntimeError('The input PauliwordOp has no Z2 symmetries.')
        if np.all(S.adjacency_matrix) or commuting_override:
            return S
        else:
            # if any of the stabilizers are not mutually commuting, take the largest commuting subset
            if S.n_terms < 10 or largest_clique:
                # expensive clique cover finding optimal commuting subset
                S_commuting = S.largest_clique(edge_relation='C')    
            else:
                # greedy graph-colouring approach when symmetry basis is large
                S_commuting = S.clique_cover(edge_relation='C', strategy='independent_set')[0]
                warnings.warn('Greedy method may identify non-optimal commuting symmetry terms; might be able to taper again.')
            
            return cls(S_commuting.symp_matrix, np.ones(S_commuting.n_terms, dtype=complex))

    def _check_stab(self) -> None:
        """ 
        Checks the stabilizer coefficients are +/-1
        """
        if not set(self.coeff_vec).issubset({0, +1,-1}):
            raise ValueError(f'Stabilizer coefficients not +/-1: {self.coeff_vec}')
        
    def _check_independent(self) -> None:
        """ 
        Check the supplied stabilizers are algebraically independent
        """
        if not check_independent(self):
            # there is a dependent row
            raise ValueError('The supplied stabilizers are not independent')

    def __str__(self) -> str:
        """ 
        Defines the print behaviour of IndependentOp - 
        returns the operator in an easily readable format

        Returns:
            out_string (str): human-readable IndependentOp string
        """
        out_string = ''
        for pauli_vec, coeff in zip(self.symp_matrix, self.coeff_vec):
            p_string = symplectic_to_string(pauli_vec)
            out_string += (f'{coeff} {p_string} \n')
        return out_string[:-2]

    def __repr__(self) -> str:
        return str(self)
    
    def __add__(self, Pword: "IndependentOp") -> "IndependentOp":
        summed = super().__add__(Pword)
        return self.from_PauliwordOp(summed)

    def _rotate_by_single_Pword(self, 
            Pword: "PauliwordOp", 
            angle: float = None
        ) -> "IndependentOp":
        rotated_stabilizers = super()._rotate_by_single_Pword(Pword, angle)
        return self.from_PauliwordOp(rotated_stabilizers)

    def perform_rotations(self, 
        rotations: List[Tuple["PauliwordOp", float]]
        ) -> "PauliwordOp":
        """ 
        Overwrite PauliwordOp.perform_rotations to return a IndependentOp.

        Args:
            rotations (List[Tuple["PauliwordOp", float]]): List of (PauliwordOp, rotation).

        Returns:
            Rotated Stabilizers of PauliwordOp Instance.
        """
        rotated_stabilizers = super().perform_rotations(rotations)
        return self.from_PauliwordOp(rotated_stabilizers)

    def _recursive_rotations(self, basis: "IndependentOp") -> None:
        """ 
        Recursively rotate terms of the IndependentOp to single-qubit Pauli operators.
        This is only possible when the basis is mutually commuting! Else, such rotations do
        not exist (there is a check for this in generate_stabilizer_rotations, that wraps this method).

        Args:
            basis (IndependentOp): Basis
        """
        # drop any term(s) that are single-qubit Pauli operators
        non_sqp = np.where(np.sum(basis.symp_matrix, axis=1)!=1)
        basis_non_sqp = IndependentOp(basis.symp_matrix[non_sqp], basis.coeff_vec[non_sqp])
        sqp_indices = np.where((basis - basis_non_sqp).symp_matrix)[1]%self.n_qubits
        self.used_indices += np.append(sqp_indices, sqp_indices+self.n_qubits).tolist()
    
        if basis_non_sqp.n_terms == 0:
            # once the basis has been fully rotated onto single-qubit Paulis, return the rotations
            return None
        else:
            # identify the lowest-weight Pauli operator from the commuting basis
            row_sum = np.sum(basis_non_sqp.symp_matrix, axis=1)
            sort_rows_by_weight = np.argsort(row_sum)
            pivot_row = basis_non_sqp.symp_matrix[sort_rows_by_weight][0]
            non_I = np.setdiff1d(np.where(pivot_row)[0], np.array(self.used_indices))
            # once a Pauli operator has been selected, the least-supported qubit is chosen as pivot
            col_sum = np.sum(basis_non_sqp.symp_matrix, axis=0)
            support = pivot_row*col_sum
            pivot_point = non_I[np.argmin(support[non_I])]
            # define (in the symplectic form) the single-qubit Pauli we aim to rotate onto
            target = np.zeros(2*self.n_qubits, dtype=int)
            target[pivot_point+self.n_qubits*(-1)**(pivot_point//self.n_qubits)]=1
            # the rotation mapping onto the target Pauli is given by (target + pivot_row)%2...
            # this is identicial to performing a bitwise XOR operation
            pivot_rotation = PauliwordOp(np.bitwise_xor(target, pivot_row), [1])
            self.stabilizer_rotations.append((pivot_rotation, None))
            # perform the rotation on the full basis (the ordering of rotations is important because of this!)
            rotated_basis = basis_non_sqp._rotate_by_single_Pword(pivot_rotation)
            return self._recursive_rotations(rotated_basis)
        
    def generate_stabilizer_rotations(self) -> None:
        """ 
        Find the full list of pi/2 Pauli rotations (Clifford operations) mapping this IndependentOp 
        to single-qubit Pauli operators, for use in stabilizer subsapce projection schemes.
        """
        assert(self.n_terms <= self.n_qubits), 'Too many terms in basis to reduce to single-qubit Paulis'
        assert(np.all(self.adjacency_matrix)), 'The basis is not commuting, hence the rotation is not possible'
        
        # ensure stabilizer_rotations and used_indices are empty before generating rotations
        self.stabilizer_rotations = []
        self.used_indices = []

        basis = self.copy()
        # generate the rotations mapping onto single-qubit Pauli operators (not necessarily the target_sqp)
        self._recursive_rotations(basis)
        rotated_basis = basis.perform_rotations(self.stabilizer_rotations)

        # now that the basis consists of single-qubit Paulis we may map to the target X,Y or Z
        for P in rotated_basis:
            # index in the X block:
            sqp_index = np.where(P.symp_matrix[0])[0][0]%self.n_qubits
            target = np.zeros(2*self.n_qubits, dtype=int)
            if self.target_sqp in ['X','Y']:
                target[sqp_index] = 1
            if self.target_sqp in ['Y','Z']:
                target[sqp_index+self.n_qubits] = 1
            R_symp = np.bitwise_xor(target, P.symp_matrix[0])
            # the rotation will be identity if already the target_sqp
            if np.any(R_symp):
                # therefore, only append nontrivial rotations
                self.stabilizer_rotations.append((PauliwordOp(R_symp, [1]),None))

    def update_sector(self, 
            ref_state: Union[List[int], np.array, QuantumState],
            threshold: float = 0.5
        ) -> None:
        """ 
        Given the specified reference state, e.g. Hartree-Fock |1...10...0>, 
        determine the corresponding sector by measuring the stabilizers.
        will also accept a superposition of basis states, in which case it will 
        identify the dominant sector therein, but note it will ascribe a zero
        assignment if there is not sufficient evidence to fix a +-1 eigenvalue.

        Args:
             ref_state (Union[List[int], np.array, QuantumState]): Reference State.
             threshold (float): Threshold Value for  expectation value. It's default value is 0.5
        """
        if not isinstance(ref_state, QuantumState):
            ref_state = QuantumState(ref_state)
        assert ref_state._is_normalized(), 'Reference state is not normalized.'
        
        ### update the stabilizers assignments in parallel
        self.coeff_vec = np.array(assign_value(self, ref_state))
        # raise a warning if any stabilizers are assigned a zero value
        if np.any(self.coeff_vec==0):
            S_zero = self[self.coeff_vec==0]; S_zero.coeff_vec[:]=1
            S_zero = list(S_zero.to_dictionary.keys())
            warnings.warn(f'The stabilizers {S_zero} were assigned zero values - bad reference state.')
        
    def rotate_onto_single_qubit_paulis(self) -> "IndependentOp":
        """ 
        Returns the rotated single-qubit Pauli stabilizers.

        Returns: 
            Rotated single-qubit Pauli stabilizers.
        """
        self.generate_stabilizer_rotations()
        if self.stabilizer_rotations != []:
            return IndependentOp.from_PauliwordOp(
                reduce(
                    lambda x,y:x.append(y), 
                    [s.perform_rotations(self.stabilizer_rotations) for s in self]
                )
            ) # ensures stabilizer ordering is correct
        else:
            return self
          
    def __getitem__(self,
                    key: Union[slice, int]
                    ) -> "IndependentOp":
        """ 
        Makes the IndependentOp subscriptable - returns a IndependentOp constructed
        from the indexed row and coefficient from the symplectic matrix.

        Args:
            key (Union[slice, int]): Key.

        Returns:
            IndependentOp constructed from the indexed row and coefficient from the symplectic matrix.
        """
        if isinstance(key, int):
            if key < 0:
                # allow negative subscript
                key += self.n_terms
            assert (key < self.n_terms), 'Index out of range'
            mask = [key]
        elif isinstance(key, slice):
            start, stop = key.start, key.stop
            if start is None:
                start = 0
            if stop is None:
                stop = self.n_terms
            mask = np.arange(start, stop, key.step)
        elif isinstance(key, (list, np.ndarray)):
            mask = np.asarray(key)
        else:
            raise ValueError('Unrecognised input, must be an integer, slice, list or np.array')

        symp_items = self.symp_matrix[mask]
        coeff_items = self.coeff_vec[mask]
        return IndependentOp(symp_items, coeff_items)

    def __iter__(self):
        """ 
        Makes a PauliwordOp instance iterable.

        Returns:
            Iterator of PauliwordOp instance.
        """
        return iter([self[i] for i in range(self.n_terms)])

@process.parallelize
def assign_value(S: PauliwordOp, ref_state: QuantumState) -> int:
    """
    Measure expectation value of stabilizer on input reference state.
    Args: 
        S (PauliwordOp): Stabilizer.
        ref_state (QuantumState): Reference State.
        threshold (float): Threshold Value of expectation value.

    Returns:
        Expectation Value (int) of stabilizer on input reference state.
    """
    threshold = 0.5
    expval = single_term_expval(S, ref_state)
    # if this expval exceeds some predefined threshold then assign the corresponding 
    # ±1 eigenvalue. Otherwise, return 0 as insufficient evidence to fix the value.
    if abs(expval) > threshold:
        return int(np.sign(expval))
    else:
        return 0



from symmer.operators import PauliwordOp, QuantumState
from symmer.projection import QubitTapering, ContextualSubspace
from symmer.utils import exact_gs_energy
from symmer.approximate import get_MPO, find_groundstate_quimb
from typing import Union, List
import numpy as np
import warnings

class QubitSubspaceManager:
    """ 
    Class for automating the following qubit subspace techqniques:

    *** QubitTapering ***
    - Maps each Z2-symmetry onto distinct qubits and projects over them
    - Yields an exact reduction, given that the correct sector is identifed

    *** ContextualSubspace ***
    - Partitions the Hamiltonian into noncontextual and contextual components
    - The noncontextual problem may be solved via classical optimization
    - We then impose noncontextual symmetries over the contextual portion of
      the Hamiltonian, constrained by the noncontextual solution
    - Allows one to define a reduced Hamiltonian of any size but incurs error
    
    It is recommended that the user should specify a reference state,
    such as Hartree-Fock. Otherwise, Symmer will try to identify an
    alternative refernce, either via direct diagonalization  if the 
    Hamiltonian is sufficiently small, or using a DMRG calculation.

    Attributes:
        _projection_ready (bool): If True, Hamiltonian is ready for projection. By default, it is set to False.
    """

    _projection_ready = False

    def __init__(self,
            hamiltonian: PauliwordOp,
            ref_state: Union[np.ndarray, List[int], QuantumState] = None,
            run_qubit_tapering: bool = True,
            run_contextual_subspace: bool = True
        ) -> None:
        """
        Args:
            hamiltonian (PauliwordOp): Hamiltonian which is to be projected.
            ref_state (QuantumState): Reference State. If no reference state is provided, then try to generate one. By default, it is set to None.
            run_qubit_tapering (bool): If True, Qubit Tapering is performed. By default, it is set to True. 
            run_contextual_subspace (bool): If True, Contextual Subspace Method is used to solve the problem. By default, it is set to True.
        """
        self.hamiltonian = hamiltonian
        self.ref_state = self.prepare_ref_state(ref_state)
        self.run_qubit_tapering = run_qubit_tapering
        self.run_contextual_subspace = run_contextual_subspace
        self.build_subspace_objects()
        
    def prepare_ref_state(self, ref_state=None) -> QuantumState:
        """ 
        If no reference state is provided, then try to generate one.
        If the Hamiltonian contains fewer than 12 qubits, we will diagonalise
        and select the true ground state. Otherwise, a cheap DMRG calculation
        will be performed to generate an approximate ground state.

        Args: 
            ref_state (QuantumState): Reference State. If no reference state is provided, then try to generate one. By default, it is set to None.

        Returns:
            ref_state (QuantumState): Generated Reference State.
        """
        if ref_state is not None:
            if isinstance(ref_state, list):
                ref_state = np.array(ref_state).reshape(-1)
            if isinstance(ref_state, np.ndarray):
                ref_state = QuantumState(ref_state, [1])
            self._aux_operator = None
        else:
            warnings.warn('No reference state supplied - trying to identify one via alternative means.')
            if self.hamiltonian.n_qubits <= 12:
                _, ref_state = exact_gs_energy(self.hamiltonian.to_sparse_matrix)
            else:
                warnings.warn(
                    'Results are currently unstable for reference state '+ 
                    'generation via tensor network techniques'
                )
                mpo = get_MPO(self.hamiltonian, max_bond_dimension=10)
                ref_state = find_groundstate_quimb(mpo)
            self._aux_operator = ref_state.state_op

        return ref_state.cleanup(zero_threshold=1e-4).normalize

    def build_subspace_objects(self) -> None:
        """ 
        Initialize the relevant qubit subspace classes.
        """
        if self.run_qubit_tapering:
            self.QT             = QubitTapering(operator=self.hamiltonian)
            self._hamiltonian   = self.QT.taper_it(ref_state=self.ref_state)
            self._ref_state     = self.QT.tapered_ref_state.normalize
            self._Z2_symmetries = self.QT.symmetry_generators.copy()
        else:
            self._hamiltonian   = self.hamiltonian.copy()
            self._ref_state     = self.ref_state.copy()
            self._Z2_symmetries = None

        if self.run_contextual_subspace:
            self.CS = ContextualSubspace(
                operator=self._hamiltonian,
                reference_state=self._ref_state,
                noncontextual_strategy='StabilizeFirst',
                noncontextual_solver  ='brute_force'
            )

    def get_reduced_hamiltonian(self, 
            n_qubits:int, aux_operator:PauliwordOp=None
        ) -> PauliwordOp:
        """ 
        Project the Hamiltonian in line with the desired qubit subspace techqniques
        and, in the case of ContextualSubspace, the desired number of qubits.

        Args:
            n_qubits (int): Number of Qubits.
            aux_operator (PauliwordOp): Auxiliary operator. By default, it is set to None.

        Returns:
            operator_out (PauliwordOp): Projection of Hamiltonian.
        """
        self._projection_ready = True
        self._n_qubits = n_qubits
        if aux_operator is None:
            aux_operator = self._aux_operator

        if n_qubits >= self.hamiltonian.n_qubits:
            warnings.warn(
                'Specified at least as many qubits as are present in the Hamiltonian - '+
                f'returning the full {self.hamiltonian.n_qubits} operator.')
            operator_out = self.hamiltonian

        elif n_qubits > self._hamiltonian.n_qubits:
            # if one wishes not to taper all the available Z2 symmetries
            assert self.run_qubit_tapering, ''
            self.QT.symmetry_generators = self._Z2_symmetries[:self.hamiltonian.n_qubits-n_qubits]
            operator_out = self.QT.taper_it(ref_state=self.ref_state)

        else:
            if self.run_qubit_tapering:
                if not self.run_contextual_subspace and n_qubits < self._hamiltonian.n_qubits:
                    warnings.warn(
                        'When contextual subspace is not run we may only reduce '+ 
                        'the Hamiltonian by the number of Z2 symmetries present. '+
                        f'The reduced Hamiltonian will contain {self._hamiltonian.n_qubits} qubits.'
                    )
                self.QT.symmetry_generators = self._Z2_symmetries
                aux_operator = self.QT.taper_it(aux_operator=aux_operator)
                operator_out = self._hamiltonian

            if self.run_contextual_subspace:
                assert n_qubits is not None, 'Must supply the desired number of qubits for the contextual subspace.'
                self.CS.update_stabilizers(
                    n_qubits=n_qubits, aux_operator=aux_operator, strategy='aux_preserving'
                )
                operator_out = self.CS.project_onto_subspace()

            if not self.run_qubit_tapering and not self.run_contextual_subspace:
                warnings.warn('Not running any subspace methods - returning the original Hamiltonian')
                operator_out = self.hamiltonian

        return operator_out
        
    def project_auxiliary_operator(self, operator: PauliwordOp) -> PauliwordOp:
        """ 
        Project additional operators consistently with respect to the Hamiltonian.

        Args:
            operator (PauliwordOp): Additional operator which has to be projected consistently with respect to the Hamiltonian.

        Returns:
            operator (PauliwordOp): Projection of additional operator.
        """
        assert self._projection_ready, 'Have not yet projected the Hamiltonian into the contextual subspace'

        if self._n_qubits < self.hamiltonian.n_qubits:

            if self.run_qubit_tapering:
                operator = self.QT.taper_it(aux_operator=operator)
            
            if self.run_contextual_subspace:
                operator = self.CS.project_onto_subspace(operator_to_project=operator)

        return operator
    
    def project_auxiliary_state(self, state: QuantumState) -> QuantumState:
        """ 
        Project quantum state consistently with respect to the Hamiltonian.

        Args:
            operator (QuantumState): Quantum State which has to be projected consistently with respect to the Hamiltonian.
        
        Returns:
            operator (PauliwordOp): Projection of Quantum State.
        """
        assert self._projection_ready, 'Have not yet projected the Hamiltonian into the contextual subspace'

        if self._n_qubits < self.hamiltonian.n_qubits:

            if self.run_qubit_tapering:
                state = self.QT.project_state(state_to_project=state)
            if self.run_contextual_subspace:
                state = self.CS.project_state(state_to_project=state)

        return state



    
    

